from snmp.dispatcher import *
from snmp.manager.v1 import *
from snmp.manager.v2c import *
from snmp.manager.v3 import *
from snmp.message import *
from snmp.message.v1 import SNMPv1MessageProcessor
from snmp.message.v2c import SNMPv2cMessageProcessor
from snmp.message.v3 import SNMPv3MessageProcessor
from snmp.security import *
from snmp.security.usm import *
from snmp.transport import *
from snmp.transport.udp import *
from snmp.typing import *

Address = Tuple[str, int]

class Engine:
    TRANSPORTS = {
        cls.DOMAIN: cls for cls in [
            UdpIPv4Socket,
            UdpIPv6Socket,
        ]
    }

    def __init__(self,
        defaultVersion: ProtocolVersion = ProtocolVersion.SNMPv3,
        defaultDomain: TransportDomain = TransportDomain.UDP_IPv4,
        defaultSecurityModel: SecurityModel = SecurityModel.USM,
        defaultCommunity: bytes = b"",
        msgMaxSize: int = 1472,
        autowait: bool = True,
    ):
        # Read-only variables
        self.defaultVersion         = defaultVersion
        self.defaultDomain          = defaultDomain
        self.defaultSecurityModel   = defaultSecurityModel
        self.defaultCommunity       = defaultCommunity
        self.autowaitDefault        = autowait

        self.msgMaxSize = msgMaxSize
        self.dispatcher = Dispatcher(UdpMultiplexor(self.msgMaxSize))
        self.transports: Dict[
            TransportDomain,
            Dict[Address, Transport[Address]]
        ] = {}

        self.mpv1: Optional[SNMPv1MessageProcessor] = None
        self.mpv2c: Optional[SNMPv2cMessageProcessor] = None
        self.mpv3: Optional[SNMPv3MessageProcessor] = None
        self._usm: Optional[UserBasedSecurityModule] = None

    @property
    def usm(self) -> UserBasedSecurityModule:
        if self._usm is None:
            self._usm = UserBasedSecurityModule()
        return self._usm

    def __enter__(self) -> "Engine":
        return self

    def __exit__(self, *args: Any) -> None:
        self.shutdown()

    def shutdown(self) -> None:
        self.dispatcher.shutdown()

    def connectTransport(self, transport: Transport[Tuple[str, int]]) -> None:
        if transport.DOMAIN in self.transports:
            errmsg = "{} is already handled by a different transport object"
            raise ValueError(errmsg.format(transport.DOMAIN))
        elif transport.DOMAIN not in self.TRANSPORTS:
            errmsg = f"Unsupported transport domain: {transport.DOMAIN}"
            raise ValueError(errmsg)

        self.dispatcher.connectTransport(transport)
        self.transports.setdefault(transport.DOMAIN, {})

    def v1Manager(self,
        channel: TransportChannel[Address],
        autowait: bool,
        community: Optional[bytes] = None,
    ) -> SNMPv1Manager[Address]:
        if community is None:
            community = self.defaultCommunity

        if self.mpv1 is None:
            self.mpv1 = SNMPv1MessageProcessor()
            self.dispatcher.addMessageProcessor(self.mpv1)

        return SNMPv1Manager(
            self.dispatcher,
            channel,
            community,
            autowait,
        )

    def v2cManager(self,
        channel: TransportChannel[Address],
        autowait: bool,
        community: Optional[bytes] = None,
    ) -> SNMPv2cManager[Address]:
        if community is None:
            community = self.defaultCommunity

        if self.mpv2c is None:
            self.mpv2c = SNMPv2cMessageProcessor()
            self.dispatcher.addMessageProcessor(self.mpv2c)

        return SNMPv2cManager(
            self.dispatcher,
            channel,
            community,
            autowait,
        )

    def v3Manager(self,
        channel: TransportChannel[Address],
        autowait: bool,
        engineID: Optional[bytes] = None,
        securityModel: Optional[SecurityModel] = None,
        defaultSecurityLevel: Optional[SecurityLevel] = None,
        **kwargs: Any,
    ) -> SNMPv3UsmManager[Address]:
        if securityModel is None:
            securityModel = self.defaultSecurityModel
        elif not isinstance(securityModel, SecurityModel):
            securityModel = SecurityModel(securityModel)

        if self.mpv3 is None:
            self.mpv3 = SNMPv3MessageProcessor(self.msgMaxSize)
            self.dispatcher.addMessageProcessor(self.mpv3)

        if securityModel == SecurityModel.USM:
            defaultUserName = kwargs.get("defaultUserName")
            namespace = kwargs.get("namespace", "")

            if defaultUserName is None:
                defaultUserName = self.usm.getDefaultUser(namespace)

            if defaultSecurityLevel is None:
                defaultSecurityLevel = self.usm.getDefaultSecurityLevel(
                    defaultUserName,
                    namespace,
                )

            self.mpv3.addSecurityModuleIfNeeded(self.usm)

            return SNMPv3UsmManager(
                self.dispatcher,
                self.usm,
                channel,
                namespace,
                defaultUserName.encode(),
                defaultSecurityLevel,
                engineID=engineID,
                autowait=autowait,
            )
        else:
            errmsg = f"Unsupported security model: {str(securityModel)}"
            raise ValueError(errmsg)

    def Manager(self,
        address: Any,
        version: Optional[ProtocolVersion] = None,
        domain: Optional[TransportDomain] = None,
        localAddress: Any = None,
        autowait: Optional[bool] = None,
        **kwargs: Any,
    ) -> Union[
        SNMPv1Manager[Address],
        SNMPv2cManager[Address],
        SNMPv3UsmManager[Address],
    ]:
        if domain is None:
            domain = self.defaultDomain

        if autowait is None:
            autowait = self.autowaitDefault

        try:
            transportClass = self.TRANSPORTS[domain]
        except KeyError as err:
            errmsg = f"Unsupported transport domain: {transportClass.DOMAIN}"
            raise ValueError(errmsg) from err

        address = transportClass.normalizeAddress(
            address,
            AddressUsage.LISTENER,
        )

        localAddress = transportClass.normalizeAddress(
            localAddress,
            AddressUsage.SENDER,
        )

        try:
            transports = self.transports[domain]
        except KeyError:
            transports = {}
            self.transports[domain] = transports

        try:
            transport = transports[localAddress]
        except KeyError:
            transport = transportClass(*localAddress)
            self.dispatcher.connectTransport(transport)
            transports[localAddress] = transport

        channel = TransportChannel(transport, address, localAddress)

        if version is None:
            version = self.defaultVersion
        elif not isinstance(version, ProtocolVersion):
            version = ProtocolVersion(version)

        if version == ProtocolVersion.SNMPv3:
            return self.v3Manager(channel, autowait, **kwargs)
        elif version == ProtocolVersion.SNMPv2c:
            return self.v2cManager(channel, autowait, **kwargs)
        elif version == ProtocolVersion.SNMPv1:
            return self.v1Manager(channel, autowait, **kwargs)
        else:
            raise ValueError(f"Unsupported protocol version: {str(version)}")



__all__ = ["EncodeError", "ParseError", "Asn1Data", "Tag", "decode", "encode"]

from enum import IntEnum

from snmp.exception import *
from snmp.typing import *
from snmp.utils import *

Asn1Data = Union[bytes, subbytes]

@final
class EncodeError(SNMPException):
    """Failure to encode a payload under ASN.1 Basic Encoding Rules."""
    pass

@final
class ParseError(IncomingMessageError):
    """Failure to translate a byte string into an ASN.1 object."""
    pass

@final
class Tag:
    """Represents an ASN.1 BER tag."""
    @final
    class Class(IntEnum):
        """Named constants for the class bits of an ASN.1 BER tag."""
        UNIVERSAL         = 0
        APPLICATION       = 1
        CONTEXT_SPECIFIC  = 2
        PRIVATE           = 3

    def __init__(self,
        number: int,
        constructed: bool = False,
        cls: Class = Class.UNIVERSAL,
    ) -> None:
        self.cls = cls
        self.constructed = constructed
        self.number = number

    def __eq__(self, other: Any) -> bool:
        if other.__class__ != self.__class__:
            return NotImplemented

        return (    # type: ignore[no-any-return]
            self.cls == other.cls
        and self.constructed == other.constructed
        and self.number == other.number
        )

    def __hash__(self) -> int:
        return hash((self.cls, self.constructed, self.number))

    def __repr__(self) -> str:
        return f"Tag({self.number}, {self.constructed}, {self.cls})"

    @classmethod
    def decode(cls, data: subbytes) -> Tuple["Tag", subbytes]:
        """Extract the tag from an ASN.1 BER string.

        This function decodes the tag portion of a BER string an returns it as
        a Tag object. It also returns a subbytes of the data argument
        referencing everything after the tag.
        """
        try:
            byte = data.dereference()
        except IndexError as err:
            raise ParseError("Missing tag") from err
        else:
            data = data.advance()

        class_      = (byte & 0xc0) >> 6
        constructed = (byte & 0x20) != 0
        number      = (byte & 0x1f)

        if number == 0x1f:
            number = 0
            byte = 0x80
            while byte & 0x80:
                try:
                    byte = data.dereference()
                except IndexError as err:
                    raise ParseError("Incomplete tag") from err
                else:
                    data = data.advance()

                number <<= 7
                number |= byte & 0x7f

        return cls(number, constructed, cls.Class(class_)), data

    def encode(self) -> bytes:
        """Encode a Tag under ASN.1 Basic Encoding Rules."""
        byte = (self.cls << 6) & 0xc0

        if self.constructed:
            byte |= 0x20

        arr = bytearray()
        if self.number < 0x1f:
            byte |= self.number
        else:
            byte |= 0x1f

            number = self.number
            indicator = 0
            while number:
                arr.append(indicator | (number & 0x7f))
                indicator = 0x80
                number >>= 7

        arr.append(byte)
        return bytes(reversed(arr))

def decode_length(data: subbytes) -> Tuple[int, subbytes]:
    """Decode the length field of an ASN.1 BER string.

    The provided data argument should contain most of a BER string, starting
    after the tag. This function will decode the length field and return it as
    an int, along with a new subbytes object referencing to the portion of the
    data argument immediately following the length field.
    """
    try:
        length = data.dereference()
    except IndexError as err:
        raise ParseError("Missing length") from err
    else:
        data = data.advance()

    if length & 0x80:
        n = length & 0x7f

        length = 0
        for i in range(n):
            try:
                byte = data.dereference()
            except IndexError as err:
                raise ParseError("Incomplete length") from err
            else:
                data = data.advance()

            length <<= 8
            length |= byte

    return length, data

def encode_length(length: int) -> bytes:
    """Encode the length of a message under ASN.1 Basic Encoding Rules."""
    if length < 0x80:
        return bytes([length])

    arr = bytearray()
    while length:
        arr.append(length & 0xff)
        length >>= 8

    if len(arr) > 0x7f:
        raise EncodeError("Length too large for definite-length encoding")

    arr.append(0x80 | len(arr))
    return bytes(reversed(arr))

def decode( # type: ignore[no-untyped-def]
    data: Asn1Data,
    expected: Optional[Tag] = None,
    leftovers: bool = False,
    copy: bool = True,
):
    """Extract the contents of an ASN.1 BER string.

    This function has several options for what it can return depending on the
    values of its arguments. In the default case, given only the data argument,
    it will return a tuple, where the first element is the Tag, and the second
    contains the body of the message as a bytes object. In this case, the
    function expects to consume the entire message, and will raise a ParseError
    if it contains more bytes than what the length field indicates. A value of
    True for the leftovers argument will change this behavior, so that it will
    return the leftover data in a snmp.utils.subbytes object as an additional
    element of the returned tuple.

    The expected argument allows the caller to provide a Tag, telling the
    function what data type is expected. In this case, the body will be
    returned without the tag, and the return value will not use a tuple (unless
    leftovers was given as True). If the decoded object does not match the
    expected type, the function will raise a ParseError.

    The copy argument allows the caller to specify whether the body of the
    message should be copied into its own bytes object (the default), or use a
    snmp.utils.subbytes object, in order to preserve a reference to the complete
    message.
    """
    data = subbytes(data)
    tag, data = Tag.decode(data)

    if expected is not None and tag != expected:
        raise ParseError("Tag does not match expected type")

    length, data = decode_length(data)
    data, tail = data.split(length)

    if len(data) < length:
        raise ParseError("Incomplete value")
    elif not leftovers and tail:
        raise ParseError("Trailing bytes")

    body = data[:] if copy else data

    if expected is None:
        if leftovers:
            return tag, body, tail
        else:
            return tag, body
    else:
        if leftovers:
            return body, tail
        else:
            return body

def encode(tag: Tag, data: bytes) -> bytes:
    """Encode a message under ASN.1 Basic Encoding Rules."""
    return tag.encode() + encode_length(len(data)) + data



__all__ = [
    "InvalidEngineID", "InvalidUserName", "InvalidSecurityLevel",
    "AuthProtocol", "PrivProtocol", "UserBasedSecurityModule",
]

import threading

from time import time
from snmp.asn1 import *
from snmp.ber import *
from snmp.exception import IncomingMessageError
from snmp.message.v3 import *
from snmp.security import *
from snmp.security.levels import *
from snmp.smi import *
from snmp.typing import *
from snmp.utils import *

class AuthProtocol:
    def __init__(self, key: bytes) -> None:
        raise NotImplementedError()

    @classmethod
    def computeKey(cls, secret: bytes) -> bytes:
        raise NotImplementedError()

    @classmethod
    def localizeKey(cls, key: bytes, engineID: bytes) -> bytes:
        raise NotImplementedError()

    @classmethod
    def localize(cls, secret: bytes, engineID: bytes) -> bytes:
        return cls.localizeKey(cls.computeKey(secret), engineID)

    @property
    def msgAuthenticationParameters(self) -> bytes:
        raise NotImplementedError()

    def sign(self, data: bytes) -> bytes:
        raise NotImplementedError()

class PrivProtocol:
    def __init__(self, key: bytes) -> None:
        raise NotImplementedError()

    def decrypt(self,
        data: bytes,
        engineBoots: int,
        engineTime: int,
        salt: bytes,
    ) -> bytes:
        raise NotImplementedError()

    def encrypt(self,
        data: bytes,
        engineBoots: int,
        engineTime: int,
    ) -> Tuple[bytes, bytes]:
        raise NotImplementedError()

class UnsupportedSecLevel(IncomingMessageError):
    pass

class NotInTimeWindow(IncomingMessageError):
    pass

class UnknownUserName(IncomingMessageError):
    pass

class UnknownEngineID(IncomingMessageError):
    pass

class WrongDigest(IncomingMessageError):
    pass

class DecryptionError(IncomingMessageError):
    pass

class InvalidEngineID(ValueError):
    pass

class InvalidUserName(ValueError):
    pass

class InvalidSecurityLevel(ValueError):
    pass

class TimeEntry:
    def __init__(self,
        engineBoots: int,
        latestBootTime: Optional[float] = None,
        authenticated: bool = False,
    ) -> None:
        if latestBootTime is None:
            latestBootTime = time()

        self.authenticated = authenticated
        self.snmpEngineBoots = engineBoots
        self.latestBootTime = latestBootTime
        self.latestReceivedEngineTime = 0

    def snmpEngineTime(self, timestamp: float) -> int:
        return int(timestamp - self.latestBootTime)

class TimeKeeper:
    MAX_ENGINE_BOOTS: ClassVar[int] = (1 << 31) - 1
    TIME_WINDOW_SIZE: ClassVar[int] = 150

    def __init__(self) -> None:
        self.lock = threading.Lock()
        self.table: Dict[bytes, TimeEntry] = {}

    def getEngineTime(self,
        engineID: bytes,
        timestamp: Optional[float] = None,
    ) -> Tuple[int, int]:
        if timestamp is None:
            timestamp = time()

        with self.lock:
            try:
                entry = self.table[engineID]
            except KeyError:
                return 0, 0

            return entry.snmpEngineBoots, entry.snmpEngineTime(timestamp)

    def update(self,
        engineID: bytes,
        msgBoots: int = 0,
        msgTime: int = 0,
        timestamp: Optional[float] = None,
    ) -> None:
        self.updateAndVerify(engineID, msgBoots, msgTime, False, timestamp)

    def updateAndVerify(self,
        engineID: bytes,
        msgBoots: int,
        msgTime: int,
        auth: bool = True,
        timestamp: Optional[float] = None,
    ) -> bool:
        if timestamp is None:
            timestamp = time()

        with self.lock:
            try:
                entry = self.table[engineID]
            except KeyError as err:
                entry = TimeEntry(msgBoots, timestamp - msgTime, auth)
                self.table[engineID] = entry

            withinTimeWindow = False
            if not entry.authenticated:
                entry.snmpEngineBoots = msgBoots
                entry.latestBootTime = timestamp - msgTime
                entry.latestReceivedEngineTime = msgTime
                withinTimeWindow = True
            elif auth:
                if msgBoots > entry.snmpEngineBoots:
                    entry.snmpEngineBoots = msgBoots
                    entry.latestBootTime = timestamp
                    entry.latestReceivedEngineTime = 0

                if msgBoots == entry.snmpEngineBoots:
                    if msgTime > entry.latestReceivedEngineTime:
                        calculatedBootTime = timestamp - msgTime
                        if calculatedBootTime < entry.latestBootTime:
                            entry.latestBootTime = calculatedBootTime

                        entry.latestReceivedEngineTime = msgTime
                        withinTimeWindow = True
                    else:
                        snmpEngineTime = entry.snmpEngineTime(timestamp)
                        difference = snmpEngineTime - msgTime
                        if difference <= self.TIME_WINDOW_SIZE:
                            withinTimeWindow = True

                if entry.snmpEngineBoots == self.MAX_ENGINE_BOOTS:
                    withinTimeWindow = False

            if auth:
                entry.authenticated = True

            return withinTimeWindow

class Credentials:
    def __init__(self,
        authProtocol: Optional[Type[AuthProtocol]] = None,
        authSecret: Optional[bytes] = None,
        privProtocol: Optional[Type[PrivProtocol]] = None,
        privSecret: Optional[bytes] = None,
        secret: bytes = b"",
    ) -> None:
        self.authProtocol = None
        self.authKey = None
        self.privProtocol = None
        self.privKey = None

        if authProtocol is None:
            self.maxSecurityLevel = noAuthNoPriv
        else:
            if privProtocol is None:
                self.maxSecurityLevel = authNoPriv
            else:
                self.maxSecurityLevel = authPriv
                self.privProtocol = privProtocol
                self.privKey = authProtocol.computeKey(privSecret or secret)

            self.authProtocol = authProtocol
            self.authKey = authProtocol.computeKey(authSecret or secret)

class LocalizedCredentials:
    def __init__(self,
        auth: Optional[AuthProtocol] = None,
        priv: Optional[PrivProtocol] = None,
    ) -> None:
        self.auth = auth
        self.priv = priv

class DiscoveredEngine:
    def __init__(self) -> None:
        self.namespace: Optional[str] = None
        self.refCount = 0

    def assign(self, namespace: str) -> Tuple[bool, bool]:
        assigned = True
        initialized = True

        if namespace != self.namespace:
            if self.refCount:
                assigned = False
            else:
                self.namespace = namespace
                initialized = False

        if assigned:
            self.refCount += 1

        return assigned, initialized

    def release(self, namespace: str) -> bool:
        assert self.namespace == namespace
        assert self.refCount > 0

        self.refCount -= 1
        return self.refCount == 0

class UserTable:
    def __init__(self) -> None:
        self.engines: Dict[bytes, Dict[bytes, LocalizedCredentials]] = {}

    def assignCredentials(self,
        engineID: bytes,
        userName: bytes,
        credentials: LocalizedCredentials,
    ) -> None:
        try:
            users = self.engines[engineID]
        except KeyError:
            users = dict()
            self.engines[engineID] = users

        users[userName] = credentials

    def getCredentials(self,
        engineID: bytes,
        userName: bytes,
    ) -> LocalizedCredentials:
        try:
            users = self.engines[engineID]
        except KeyError as err:
            raise InvalidEngineID(engineID) from err

        try:
            return users[userName]
        except KeyError as err:
            raise InvalidUserName(userName) from err

class UserEntry:
    def __init__(self,
        defaultSecurityLevel: SecurityLevel,
        credentials: Credentials
    ) -> None:
        self.credentials = credentials
        self.defaultSecurityLevel = defaultSecurityLevel

class NameSpace:
    def __init__(self, defaultUserName: str):
        self.defaultUserName = defaultUserName
        self.users: Dict[str, UserEntry] = {}

    def __iter__(self) -> Iterator[Tuple[str, UserEntry]]:
        return self.users.items().__iter__()

    def __contains__(self, key: str) -> bool:
        return self.users.__contains__(key)

    def __getitem__(self, key: str) -> UserEntry:
        return self.users.__getitem__(key)

    def __setitem__(self, key: str, item: UserEntry) -> None:
        return self.users.__setitem__(key, item)

@final
class UsmSecurityParameters(Sequence):
    def __init__(self,
        engineID: bytes,
        engineBoots: int,
        engineTime: int,
        userName: bytes,
        signature: Asn1Data,
        salt: bytes,
    ):
        self.engineID = engineID
        self.engineBoots = engineBoots
        self.engineTime = engineTime
        self.userName = userName
        self.salt = salt

        self.signatureIndex: Optional[int]
        self.wholeMsg: Optional[bytes]

        if isinstance(signature, subbytes):
            self.signature = signature[:]
            self.signatureIndex = signature.start
            self.wholeMsg = signature.data
        else:
            self.signature = signature
            self.signatureIndex = None
            self.wholeMsg = None

    def __iter__(self) -> Iterator[ASN1]:
        yield OctetString(self.engineID)
        yield Integer(self.engineBoots)
        yield Integer(self.engineTime)
        yield OctetString(self.userName)
        yield OctetString(self.signature)
        yield OctetString(self.salt)

    def __len__(self) -> int:
        return 6

    def __repr__(self) -> str:
        args = (
            str(self.engineID),
            str(self.engineBoots),
            str(self.engineTime),
            str(self.userName),
            str(self.signature),
            str(self.salt),
        )

        return f"{typename(self)}({', '.join(args)})"

    def __str__(self) -> str:
        return self.toString()

    def toString(self, depth: int = 0, tab: str = "    ") -> str:
        indent = tab * depth
        subindent = indent + tab
        return "\n".join((
            f"{indent}{typename(self)}:",
            f"{subindent}Authoritative Engine ID: {self.engineID!r}",
            f"{subindent}Authoritative Engine Boots: {self.engineBoots}",
            f"{subindent}Authoritative Engine Time: {self.engineTime}",
            f"{subindent}User Name: {self.userName!r}",
            f"{subindent}Signature: {self.signature!r}",
            f"{subindent}Encryption Salt: {self.salt!r}",
        ))

    @overload
    @classmethod
    def decode(cls,
        data: Asn1Data,
    ) -> "UsmSecurityParameters":
        ...

    @overload
    @classmethod
    def decode(cls,
        data: Asn1Data,
        leftovers: bool = False,
        copy: bool = True,
        **kwargs: Any,
    ) -> Union[
        "UsmSecurityParameters",
        Tuple["UsmSecurityParameters", subbytes],
    ]:
        ...

    @classmethod
    def decode(cls,
        data: Asn1Data,
        leftovers: bool = False,
        copy: bool = False,
        **kwargs: Any,
    ) -> Union[
        "UsmSecurityParameters",
        Tuple["UsmSecurityParameters", subbytes],
    ]:
        return super().decode(data, leftovers, copy, **kwargs)

    @classmethod
    def deserialize(cls, data: Asn1Data) -> "UsmSecurityParameters":
        copy = not isinstance(data, subbytes)

        engineID, ptr = cast(
            Tuple[OctetString, subbytes],
            OctetString.decode(data, leftovers=True),
        )

        engineBoots, ptr = cast(
            Tuple[Integer, subbytes],
            Integer.decode(ptr, leftovers=True),
        )

        engineTime, ptr = cast(
            Tuple[Integer, subbytes],
            Integer.decode(ptr, leftovers=True),
        )

        userName, ptr = cast(
            Tuple[OctetString, subbytes],
            OctetString.decode(ptr, leftovers=True),
        )

        signature, ptr = cast(
            Tuple[OctetString, subbytes],
            OctetString.decode(ptr, leftovers=True, copy=copy),
        )

        salt = OctetString.decode(ptr)

        return cls(
            engineID.data,
            engineBoots.value,
            engineTime.value,
            userName.data,
            signature.original,
            salt.data,
        )

    @classmethod
    def findSignature(self, msgSecurityParameters: subbytes) -> subbytes:
        ptr = cast(
            subbytes,
            decode(msgSecurityParameters, Sequence.TAG, copy=False)
        )

        _, ptr = decode(ptr, OctetString.TAG,   leftovers=True, copy=False)
        _, ptr = decode(ptr, Integer.TAG,       leftovers=True, copy=False)
        _, ptr = decode(ptr, Integer.TAG,       leftovers=True, copy=False)
        _, ptr = decode(ptr, OctetString.TAG,   leftovers=True, copy=False)
        ptr, _ = decode(ptr, OctetString.TAG,   leftovers=True, copy=False)
        return ptr

class UserBasedSecurityModule(SecurityModule[SNMPv3Message]):
    MODEL = SecurityModel.USM

    def __init__(self, engineID: Optional[bytes] = None) -> None:
        self.engineID = engineID
        self.engines: Dict[bytes, DiscoveredEngine] = {}
        self.lock = threading.Lock()
        self.namespaces: Dict[str, NameSpace] = {}
        self.timekeeper = TimeKeeper()
        self.users = UserTable()

        if self.engineID is not None:
            self.timekeeper.update(self.engineID)

    @staticmethod
    def localizeCredentials(
        engineID: bytes,
        credentials: Optional[Credentials] = None,
    ) -> LocalizedCredentials:
        if credentials is None:
           credentials = Credentials()

        auth = None
        priv = None

        if credentials.authProtocol is not None:
            assert credentials.authKey is not None
            auth = credentials.authProtocol(
                credentials.authProtocol.localizeKey(
                    credentials.authKey,
                    engineID,
                )
            )

            if credentials.privProtocol is not None:
                assert credentials.privKey is not None
                priv = credentials.privProtocol(
                    credentials.authProtocol.localizeKey(
                        credentials.privKey,
                        engineID,
                    )
                )

        return LocalizedCredentials(auth, priv)

    def addUser(self,
        userName: str,
        authProtocol: Optional[Type[AuthProtocol]] = None,
        authSecret: Optional[bytes] = None,
        privProtocol: Optional[Type[PrivProtocol]] = None,
        privSecret: Optional[bytes] = None,
        secret: bytes = b"",
        default: bool = False,
        defaultSecurityLevel: Optional[SecurityLevel] = None,
        namespace: str = "",
    ) -> None:
        credentials = Credentials(
            authProtocol,
            authSecret,
            privProtocol,
            privSecret,
            secret,
        )

        if defaultSecurityLevel is None:
            defaultSecurityLevel = credentials.maxSecurityLevel
        elif defaultSecurityLevel > credentials.maxSecurityLevel:
            errmsg = "Unable to support {} without the \"{}\" argument"

            if credentials.maxSecurityLevel.auth:
                param = "privProtocol"
            else:
                param = "authProtocol"

            raise ValueError(errmsg.format(defaultSecurityLevel, param))

        with self.lock:
            try:
                space = self.namespaces[namespace]
            except KeyError:
                space = NameSpace(userName)
                self.namespaces[namespace] = space
            else:
                if userName in space:
                    errmsg = f"User \"{userName}\" is already defined"

                    if namespace:
                        errmsg += f" in namespace \"{namespace}\""

                    raise ValueError(errmsg)

                if default:
                    space.defaultUserName = userName

            space[userName] = UserEntry(defaultSecurityLevel, credentials)

    def getDefaultSecurityLevel(self,
        userName: str,
        namespace: str = "",
    ) -> SecurityLevel:
        space = self.getNameSpace(namespace)

        try:
            user = space[userName]
        except KeyError as err:
            errmsg = f"No user \"{userName}\""

            if namespace:
                errmsg += f" in namespace \"{namespace}\""

            raise ValueError(errmsg) from err

        return user.defaultSecurityLevel

    def getDefaultUser(self, namespace: str = "") -> str:
        return self.getNameSpace(namespace).defaultUserName

    def getNameSpace(self, namespace: str = "") -> NameSpace:
        try:
            return self.namespaces[namespace]
        except KeyError as err:
            errmsg = "No users defined"

            if namespace:
                errmsg += f" in namespace \"{namespace}\""

            raise ValueError(errmsg) from err

    def registerRemoteEngine(self,
        engineID: bytes,
        namespace: str = "",
    ) -> bool:
        with self.lock:
            try:
                engine = self.engines[engineID]
            except KeyError:
                engine = DiscoveredEngine()
                self.engines[engineID] = engine

            assigned, initialized = engine.assign(namespace)

            # Read as "assigned but not initialized"
            if not initialized and assigned:
                ns = self.namespaces[namespace]
                for userName, entry in ns:
                    self.users.assignCredentials(
                        engineID,
                        userName.encode(),
                        self.localizeCredentials(engineID, entry.credentials)
                    )

            return assigned

    def unregisterRemoteEngine(self,
        engineID: bytes,
        namespace: str = "",
    ) -> None:
        with self.lock:
            try:
                engine = self.engines[engineID]
            except KeyError:
                assert False, f"Engine {engineID!r} was never registered"
            else:
                if engine.release(namespace):
                    del self.engines[engineID]

    def prepareOutgoing(self,
        message: SNMPv3Message,
        engineID: bytes,
        securityName: bytes,
        timestamp: Optional[float] = None,
    ) -> bytes:
        if timestamp is None:
            timestamp = time()

        if message.header.flags.authFlag:
            with self.lock:
                user = self.users.getCredentials(engineID, securityName)

            if not user.auth:
                userName = securityName.decode
                errmsg = f"Authentication is disabled for user \"{userName}\""
                raise InvalidSecurityLevel(errmsg)

            snmpEngineBoots, snmpEngineTime = self.timekeeper.getEngineTime(
                engineID,
                timestamp=timestamp,
            )

            msgAuthenticationParameters = user.auth.msgAuthenticationParameters
            msgPrivacyParameters = b''

            if message.header.flags.privFlag:
                if not user.priv:
                    userName = securityName.decode
                    errmsg = f"Privacy is disabled for user \"{userName}\""
                    raise InvalidSecurityLevel(errmsg)

                ciphertext, msgPrivacyParameters = user.priv.encrypt(
                    cast(ScopedPDU, message.scopedPDU).encode(),
                    snmpEngineBoots,
                    snmpEngineTime,
                )

                message.encryptedPDU = OctetString(ciphertext)

        else:
            if engineID == self.engineID:
                engineTimeParameters = self.timekeeper.getEngineTime(
                    engineID,
                    timestamp=timestamp,
                )

                snmpEngineBoots, snmpEngineTime = engineTimeParameters
            else:
                snmpEngineBoots = 0
                snmpEngineTime = 0

            msgAuthenticationParameters = b''
            msgPrivacyParameters = b''

        securityParameters = UsmSecurityParameters(
            engineID,
            snmpEngineBoots,
            snmpEngineTime,
            securityName,
            msgAuthenticationParameters,
            msgPrivacyParameters,
        )

        message.securityParameters = OctetString(securityParameters.encode())
        wholeMsg = message.encode()

        if message.header.flags.authFlag:
            location = UsmSecurityParameters.findSignature(
                SNMPv3Message.findSecurityParameters(wholeMsg)
            )

            signature = cast(AuthProtocol, user.auth).sign(wholeMsg)
            wholeMsg = location.replace(signature)

        return wholeMsg

    def processIncoming(self,
        message: SNMPv3Message,
        timestamp: Optional[float] = None,
    ) -> None:
        if timestamp is None:
            timestamp = time()

        securityParameters = UsmSecurityParameters.decode(
            message.securityParameters.original,
        )

        message.securityEngineID = securityParameters.engineID
        message.securityName     = securityParameters.userName

        remoteIsAuthoritative = (securityParameters.engineID != self.engineID)

        if not message.header.flags.authFlag:
            if remoteIsAuthoritative:
                self.timekeeper.update(
                    securityParameters.engineID,
                    securityParameters.engineBoots,
                    securityParameters.engineTime,
                    timestamp=timestamp
                )

            return

        try:
            with self.lock:
                user = self.users.getCredentials(
                    securityParameters.engineID,
                    securityParameters.userName,
                )
        except InvalidEngineID as err:
            raise UnknownEngineID(securityParameters.engineID) from err
        except InvalidUserName as err:
            raise UnknownUserName(securityParameters.userName) from err

        if user.auth is None:
            username = securityParameters.userName.decode()
            errmsg = f"Authentication is disabled for user \"{username}\""
            raise UnsupportedSecLevel(errmsg)
        elif message.header.flags.privFlag and user.priv is None:
            username = securityParameters.userName.decode()
            errmsg = f"Data privacy is disabled for user \"{username}\""
            raise UnsupportedSecLevel(errmsg)

        padding = user.auth.msgAuthenticationParameters
        if len(securityParameters.signature) != len(padding):
            raise WrongDigest("Invalid signature length")

        assert securityParameters.wholeMsg is not None
        wholeMsg = bytearray(securityParameters.wholeMsg)

        assert securityParameters.signatureIndex is not None
        start = securityParameters.signatureIndex
        stop = start + len(padding)
        wholeMsg[start:stop] = padding

        if user.auth.sign(wholeMsg) != securityParameters.signature:
            raise WrongDigest("Invalid signature")

        try:
            if not self.timekeeper.updateAndVerify(
                securityParameters.engineID,
                securityParameters.engineBoots,
                securityParameters.engineTime,
                timestamp=timestamp,
            ):
                raise NotInTimeWindow((
                    securityParameters.engineID,
                    securityParameters.engineBoots,
                    securityParameters.engineTime,
                ))
        except InvalidEngineID as err:
            raise UnknownEngineID(securityParameters.engineID) from err

        if message.header.flags.privFlag:
            try:
                message.plaintext = cast(PrivProtocol, user.priv).decrypt(
                    cast(OctetString, message.encryptedPDU).data,
                    securityParameters.engineBoots,
                    securityParameters.engineTime,
                    securityParameters.salt,
                )
            except ValueError as err:
                raise DecryptionError(str(err)) from err



__all__ = [
    "Any",
    "Callable",
    "ClassVar",
    "Deque",
    "Dict",
    "Generic",
    "Iterable",
    "Iterator",
    "List",
    "Literal",
    "Mapping",
    "NamedTuple",
    "Optional",
    "Set",
    "Tuple",
    "Type",
    "TypeVar",
    "Union",
    "cast",
    "final",
    "overload",
]

from typing import Any
from typing import Callable
from typing import ClassVar
from typing import Generic
from typing import NamedTuple
from typing import Optional
from typing import TypeVar
from typing import Union
from typing import cast
from typing import overload

import sys
if sys.version_info[:2] >= (3, 9):
    from collections import deque as Deque
    from builtins import dict as Dict
    from collections.abc import Iterable
    from collections.abc import Iterator
    from builtins import list as List
    from collections.abc import Mapping
    from builtins import set as Set
    from builtins import tuple as Tuple
    from builtins import type as Type
else:
    from typing import Deque
    from typing import Dict
    from typing import Iterable
    from typing import Iterator
    from typing import List
    from typing import Mapping
    from typing import Set
    from typing import Tuple
    from typing import Type

if sys.version_info[:2] >= (3, 8):
    from typing import Literal
    from typing import final
else:
    class DummyType:
        def __getitem__(self, key):
            return None

    Literal = DummyType()

    def final(wrapped):
        return wrapped



__all__ = [
    "AddressUsage", "Transport", "TransportChannel",
    "TransportDomain", "TransportListener", "TransportMultiplexor"
]

from collections import namedtuple
import enum
import os
import socket

from snmp.typing import *

AddressUsage = enum.Enum(
    "AddressUsage",
    ["LISTENER", "TRAP_LISTENER", "SENDER"]
)

class TransportDomain(enum.Enum):
    def __init__(self, family: socket.AddressFamily, loopback: str, default: str):
        self.address_family = family
        self.loopback_address = loopback
        self.default_address = default

    UDP_IPv4 = socket.AF_INET, "127.0.0.1", "0.0.0.0"
    UDP_IPv6 = socket.AF_INET6, "::1", "::"

T = TypeVar("T")
class Transport(Generic[T]):
    DOMAIN: ClassVar[TransportDomain]

    @classmethod
    def normalizeAddress(cls,
        address: Any = None,
        usage: Optional[AddressUsage] = None,
    ) -> T:
        raise NotImplementedError()

    def close(self) -> None:
        raise NotImplementedError()

    def send(self, address: T, data: bytes) -> None:
        raise NotImplementedError()

class TransportChannel(Generic[T]):
    def __init__(self,
        transport: Transport[T],
        address: T,
        localAddress: T,
    ) -> None:
        self.transport = transport
        self.address = address
        self.localAddress = localAddress

    @property
    def domain(self) -> TransportDomain:
        return self.transport.DOMAIN

    def send(self, data: bytes) -> None:
        self.transport.send(self.address, data)

class TransportListener(Generic[T]):
    def hear(self, transport: Transport[T], address: T, data: bytes) -> None:
        raise NotImplementedError()

class TransportMultiplexor(Generic[T]):

    def register(self, sock: Transport[T]) -> None:
        raise NotImplementedError()

    def listen(self, listener: TransportListener[T]) -> None:
        raise NotImplementedError()

    def stop(self) -> None:
        raise NotImplementedError()

    def close(self) -> None:
        raise NotImplementedError()

supported = ("posix")
if os.name in supported:
    package = os.name
else:
    package = "generic"

package = "{}.{}".format(__name__, package)



class SNMPException(Exception):
    """Base class for all run-time exceptions in this library."""
    pass

class SNMPLibraryBug(AssertionError):
    """Base class for logic errors in the code of this library."""
    pass

class UnsupportedFeature(SNMPLibraryBug):
    """A failure owing to a feature that this library does not yet support."""
    pass

class IncomingMessageError(SNMPException):
    """An error indicating a received message is invalid in some way."""
    pass



__all__ = [
    "SecurityLevel", "SecurityModel", "SecurityModule", "SecurityParameters",
]

import enum

from snmp.smi import *
from snmp.typing import *
from snmp.utils import *

TMessage = TypeVar("TMessage", bound="Sequence")

class SecurityLevel:
    def __init__(self, auth: Any = False, priv: Any = False) -> None:
        a = bool(auth)
        p = bool(priv)

        if p and not a:
            raise ValueError("Privacy without authentication is not valid")

        self._auth = a
        self._priv = p

    def __repr__(self) -> str:
        return "{}(auth={}, priv={})".format(
            typename(self),
            self.auth,
            self.priv
        )

    def __str__(self) -> str:
        return "{}{}".format(
            "auth" if self.auth else "noAuth",
            "Priv" if self.priv else "NoPriv"
        )

    @property
    def auth(self) -> bool:
        return self._auth

    @property
    def priv(self) -> bool:
        return self._priv

    def __eq__(self, other: Any) -> bool:
        try:
            result = (self.auth == other.auth and self.priv == other.priv)
        except AttributeError:
            return NotImplemented
        else:
            return cast(bool, result)

    def __lt__(self, other: "SecurityLevel") -> bool:
        if self.auth:
            return other.priv and not self.priv
        else:
            return other.auth

    def __ge__(self, other: "SecurityLevel") -> bool:
        return not self < other

class SecurityModel(enum.IntEnum):
    USM = 3

class SecurityParameters:
    def __init__(self, engineID: bytes, userName: bytes):
        self.securityEngineID = engineID
        self.securityName = userName

    def __repr__(self) -> str:
        return "{}({!r}, {!r})".format(
            typename(self),
            self.securityEngineID,
            self.securityName,
        )

class SecurityModule(Generic[TMessage]):
    MODEL: ClassVar[SecurityModel]

    def processIncoming(self,
        message: TMessage,
        timestamp: Optional[float] = None,
    ) -> None:
        raise NotImplementedError()

    def prepareOutgoing(self,
        message: TMessage,
        engineID: bytes,
        securityName: bytes,
    ) -> bytes:
        raise NotImplementedError()



import threading
import weakref

from snmp.ber import *
from snmp.exception import *
from snmp.message import *
from snmp.pdu import *
from snmp.typing import *
from snmp.utils import *

pduTypes = {
    cls.TAG: cls for cls in cast(Tuple[Type[AnyPDU], ...], (
        GetRequestPDU,
        GetNextRequestPDU,
        ResponsePDU,
        SetRequestPDU,
        GetBulkRequestPDU,
        InformRequestPDU,
        SNMPv2TrapPDU,
        ReportPDU,
    ))
}

class CacheError(SNMPException):
    pass

class LateResponse(IncomingMessageError):
    pass

class ResponseMismatch(IncomingMessageError):
    pass

class CacheEntry:
    def __init__(self,
        handle: RequestHandle[Message],
        community: bytes,
    ) -> None:
        self.community = community
        self.handle = weakref.ref(handle)

class SNMPv2cMessageProcessor(MessageProcessor[Message, AnyPDU]):
    VERSION = ProtocolVersion.SNMPv2c

    def __init__(self) -> None:
        self.cacheLock = threading.Lock()
        self.generator = self.newGenerator()
        self.outstanding: Dict[int, CacheEntry] = {}

    @staticmethod
    def newGenerator() -> NumberGenerator:
        return NumberGenerator(32)

    def cache(self, entry: CacheEntry) -> int:
        retry = 0
        while retry < 10:
            with self.cacheLock:
                requestID = next(self.generator)
                if requestID == 0:
                    self.generator = self.newGenerator()
                elif requestID not in self.outstanding:
                    self.outstanding[requestID] = entry
                    return requestID

            retry += 1

        raise CacheError("Failed to allocate request ID")

    def retrieve(self, requestID: int) -> CacheEntry:
        with self.cacheLock:
            return self.outstanding[requestID]

    def uncache(self, requestID: int) -> None:
        with self.cacheLock:
            try:
                del self.outstanding[requestID]
            except KeyError:
                pass

    def prepareDataElements(self,
        msg: Asn1Data,
    ) -> Tuple[Message, RequestHandle[Message]]:
        message = cast(Message, Message.decode(msg, types=pduTypes))

        if isinstance(message.pdu, ResponsePDU):
            try:
                entry = self.retrieve(message.pdu.requestID)
            except KeyError as err:
                errmsg = f"Unknown requestID: {message.pdu.requestID}"
                raise ResponseMismatch(errmsg) from err

            handle = entry.handle()
            if handle is None:
                self.uncache(message.pdu.requestID)
                raise LateResponse("Handle has already been released")

            if entry.community != message.community:
                raise ResponseMismatch(f"Community does not match request")
        else:
            raise UnsupportedFeature("Received a non-response PDU type")

        return message, handle

    def prepareOutgoingMessage(self,
        pdu: AnyPDU,
        handle: RequestHandle[Message],
        community: bytes,
    ) -> bytes:
        if pdu.requestID == 0:
            cacheEntry = CacheEntry(handle, community)
            pdu.requestID = self.cache(cacheEntry)
            handle.addCallback(self.uncache, pdu.requestID)

        return Message(self.VERSION, community, pdu).encode()



__all__ = [
    "Integer", "Integer32", "Unsigned", "Unsigned32",
    "Counter32", "Gauge32", "TimeTicks", "Counter64",
    "OctetString", "IpAddress", "Opaque",
    "NULL", "Null", "OID", "zeroDotZero",
    "Sequence",
]

from socket import inet_aton, inet_ntoa

from snmp.asn1 import *
from snmp.ber import *
from snmp.typing import *
from snmp.utils import *

TPrimitive      = TypeVar("TPrimitive",     bound="Primitive")
TInteger        = TypeVar("TInteger",       bound="BoundedInteger")
TOctetString    = TypeVar("TOctetString",   bound="OctetString")

class BoundedInteger(INTEGER):
    BITS:   ClassVar[int]
    SIGNED: ClassVar[bool]

    def __init__(self, value: int) -> None:
        if not self.inRange(value):
            raise ValueError(f"{value} is out of range for {typename(self)}")

        super().__init__(value)

    @classmethod
    def construct(cls: Type[TInteger], value: int) -> TInteger:
        try:
            return cls(value)
        except ValueError as err:
            raise ParseError(*err.args)

    @classmethod
    def inRange(cls, value: int) -> bool:
        if value < 0 and not cls.SIGNED:
            return False

        nbits = cls.bitCount(value)
        allowable = cls.BITS - 1 if cls.SIGNED else cls.BITS
        return nbits <= allowable

class Integer32(BoundedInteger):
    BITS = 32
    SIGNED  = True

class Unsigned32(BoundedInteger):
    BITS = 32
    SIGNED = False
    TAG = Tag(2, cls = Tag.Class.APPLICATION)

Integer = Integer32
Unsigned = Unsigned32

@final
class Counter32(Unsigned32):
    TAG = Tag(1, cls = Tag.Class.APPLICATION)

@final
class Gauge32(Unsigned32):
    pass

@final
class TimeTicks(Unsigned32):
    TAG = Tag(3, cls = Tag.Class.APPLICATION)

@final
class Counter64(BoundedInteger):
    BITS = 64
    SIGNED = False
    TAG = Tag(6, cls = Tag.Class.APPLICATION)

class OctetString(OCTET_STRING):
    def __init__(self, data: Asn1Data = b"") -> None:
        self.check(data)

        self._original = data
        if isinstance(data, subbytes):
            data = data[:]

        super().__init__(data)

    def __repr__(self) -> str:
        return f"{typename(self)}({self.original!r})"

    @property
    def original(self) -> Asn1Data:
        return self._original

    @classmethod
    def check(cls, data: Asn1Data) -> None:
        if len(data) > 0xffff:
            raise ValueError(f"{typename(cls)} is limited to {0xffff} bytes")

    @classmethod
    def construct(cls: Type[TOctetString], data: Asn1Data) -> TOctetString:
        try:
            return cls(data)
        except ValueError as err:
            raise ParseError(*err.args) from err

@final
class IpAddress(OCTET_STRING):
    TAG = Tag(0, cls = Tag.Class.APPLICATION)

    def __init__(self, addr: str) -> None:
        try:
            data = inet_aton(addr)
        except OSError as err:
            raise ValueError(f"Invalid IPv4 address: \"{addr}\"") from err

        super().__init__(data)
        self.addr = addr

    def __repr__(self) -> str:
        return f"{typename(self)}({repr(self.addr)})"

    def asOID(self, implied: bool = False) -> Iterable[int]:
        return super().asOID(implied=True)

    @classmethod
    def construct(cls, data: Asn1Data) -> "IpAddress":
        if isinstance(data, subbytes):
            data = data[:]

        try:
            addr = inet_ntoa(data)
        except OSError as err:
            raise ParseError(f"Invalid IPv4 address: {data!r}") from err

        return cls(addr)

@final
class Opaque(OCTET_STRING):
    TAG = Tag(4, cls = Tag.Class.APPLICATION)

Null = NULL

class OID(OBJECT_IDENTIFIER):
    def __init__(self, *subidentifiers: int) -> None:
        if len(subidentifiers) > 128:
            errmsg = "OID may not contain more than 128 sub-identifiers"
            raise ValueError(errmsg)

        if any(map(lambda x: x.bit_length() > 32, subidentifiers)):
            raise ValueError("Sub-identifiers are limited to 32-bits unsigned")

        super().__init__(*subidentifiers)

    def getIndex(self,
        prefix: "OID",
        cls: Type[TPrimitive] = Integer,    # type: ignore[assignment]
        implied: bool = False,
    ) -> TPrimitive:
        return self.decodeIndex(prefix, cls, implied=implied)[0]

zeroDotZero = OID(0, 0)

class Sequence(SEQUENCE):
    pass



__all__ = ["UdpMultiplexor"]

import select
from socket import *

from snmp.transport import *
from snmp.transport.udp import UdpSocket
from snmp.typing import *

STOPMSG = bytes(1)

class GenericUdpMultiplexor(TransportMultiplexor[Tuple[str, int]]):
    def __init__(self, recvSize: int = 1472) -> None:
        self.recvSize = recvSize
        self.r: Optional[socket] = None
        self.w: Optional[socket] = None
        self.sockets: Dict[int, UdpSocket] = {}

    def register(self, sock: UdpSocket) -> None:    # type: ignore[override]
        if self.w is None:
            address_family = sock.DOMAIN.address_family
            loopback_address = sock.DOMAIN.loopback_address

            self.r = socket(address_family, SOCK_DGRAM)
            self.w = socket(address_family, SOCK_DGRAM)

            self.r.setblocking(False)
            self.r.bind((loopback_address, 0))
            self.w.bind((loopback_address, 0))

        self.sockets[sock.fileno] = sock

    def listen(self, listener: TransportListener[Tuple[str, int]]) -> None:
        readfds = []

        if self.r is not None:
            readfds.append(self.r.fileno())

        for fileno in self.sockets.keys():
            readfds.append(fileno)

        done = not readfds
        while not done:
            ready = select.select(readfds, [], [])[0]
            for fd in ready:
                try:
                    sock = self.sockets[fd]
                except KeyError:
                    assert self.r is not None
                    assert self.w is not None
                    if fd == self.r.fileno():
                        data, addr = self.r.recvfrom(len(STOPMSG))
                        done = addr == self.w.getsockname() and data == STOPMSG
                else:
                    addr, data = sock.receive(self.recvSize)
                    listener.hear(sock, addr, data)

    def stop(self) -> None:
        if self.w is not None:
            assert self.r is not None
            self.w.sendto(STOPMSG, self.r.getsockname())

    def close(self) -> None:
        for sock in self.sockets.values():
            sock.close()

        if self.w is not None:
            assert self.r is not None
            self.w.close()
            self.r.close()

UdpMultiplexor = GenericUdpMultiplexor



__all__ = ["DesCbc"]

import os

from Crypto.Cipher import DES
from Crypto.Cipher._mode_cbc import CbcMode

from snmp.security.usm import DecryptionError, PrivProtocol
from snmp.typing import *

class DesCbc(PrivProtocol):
    BYTEORDER:  ClassVar[Literal["big"]] = "big"

    BLOCKLEN:   ClassVar[int] = 8
    KEYLEN:     ClassVar[int] = BLOCKLEN * 2
    SALTLEN:    ClassVar[int] = BLOCKLEN // 2
    SALTWRAP:   ClassVar[int] = 1 << (8 * SALTLEN)

    def __init__(self, key: bytes) -> None:
        if len(key) < self.KEYLEN:
            errmsg = f"key must be at least {self.KEYLEN} bytes long"
            raise ValueError(errmsg)

        self.key = key[:self.BLOCKLEN]
        self.preIV = key[self.BLOCKLEN:self.KEYLEN]
        self.salt = int.from_bytes(os.urandom(self.SALTLEN), self.BYTEORDER)

    def computeIV(self, salt: bytes) -> bytes:
        return bytes(a ^ b for a, b in zip(self.preIV, salt))

    def newCipher(self, iv: bytes) -> CbcMode:
        return cast(CbcMode, DES.new(self.key, DES.MODE_CBC, iv=iv))

    def pad(self, data: bytes) -> bytes:
        n = self.BLOCKLEN - (len(data) % self.BLOCKLEN)
        return data + bytes(n)

    def decrypt(self,
        data: bytes,
        engineBoots: int,
        engineTime: int,
        salt: bytes,
    ) -> bytes:
        if len(data) % self.BLOCKLEN:
            errmsg = "DES ciphertext must be a multiple of {} in length"
            raise DecryptionError(errmsg.format(self.BLOCKLEN))

        return self.newCipher(self.computeIV(salt)).decrypt(self.pad(data))

    def encrypt(self,
        data: bytes,
        engineBoots: int,
        engineTime: int,
    ) -> Tuple[bytes, bytes]:
        self.salt = (self.salt + 1) % self.SALTWRAP
        salt = b''.join((
            engineBoots.to_bytes(self.BLOCKLEN - self.SALTLEN, self.BYTEORDER),
            self.salt.to_bytes(self.SALTLEN, self.BYTEORDER),
        ))

        iv = self.computeIV(salt)
        ciphertext = self.newCipher(iv).encrypt(self.pad(data))

        return ciphertext, salt



__all__ = ["Dispatcher"]

import threading

from snmp.ber import ParseError, decode
from snmp.exception import *
from snmp.message import *
from snmp.pdu import AnyPDU
from snmp.security.levels import noAuthNoPriv
from snmp.transport import *
from snmp.typing import *
from snmp.utils import typename

T = TypeVar("T")
class Dispatcher(TransportListener[T]):
    def __init__(self,
        multiplexor: TransportMultiplexor[T],
    ) -> None:
        self.lock = threading.Lock()
        self.msgProcessors: Dict[
            ProtocolVersion,
            MessageProcessor[Any, Any],
        ] = {}

        self.multiplexor = multiplexor
        self.thread: Optional[threading.Thread] = None

    def addMessageProcessor(self, mp: MessageProcessor[Any, Any]) -> None:
        with self.lock:
            self.msgProcessors[mp.VERSION] = mp

    def connectTransport(self, transport: Transport[T]) -> None:
        domain = transport.DOMAIN

        if self.thread is not None:
            self.multiplexor.stop()
            self.thread.join()

        self.multiplexor.register(transport)
        self.thread = threading.Thread(
            target=self.multiplexor.listen,
            args=(self,)
        )

        self.thread.start()

    def hear(self, transport: Transport[T], address: T, data: bytes) -> None:
        try:
            try:
                msgVersion = VersionOnlyMessage.decode(data).version
            except ParseError:
                return

            with self.lock:
                try:
                    mp = self.msgProcessors[msgVersion]
                except KeyError:
                    return

            try:
                message, handle = mp.prepareDataElements(data)
                handle.push(message)
            except IncomingMessageError:
                return

        except AssertionError:
            pass
        except Exception:
            pass

    def sendPdu(self,
        channel: TransportChannel[T],
        msgVersion: ProtocolVersion,
        pdu: AnyPDU,
        handle: RequestHandle,  # type: ignore[type-arg]
        *args: Any,
        **kwargs: Any,
    ) -> None:
        with self.lock:
            try:
                mp = self.msgProcessors[msgVersion]
            except KeyError as err:
                version = str(ProtocolVersion(msgVersion))
                raise ValueError("{} is not enabled".format(version)) from err

        msg = mp.prepareOutgoingMessage(pdu, handle, *args, **kwargs)
        channel.send(msg)

    def shutdown(self) -> None:
        if self.thread is not None:
            self.multiplexor.stop()
            self.thread.join()

        self.multiplexor.close()
        self.thread = None



__all__ = [
    "BadVersion", "ProtocolVersion",
    "VersionOnlyMessage", "Message",
    "RequestHandle", "MessageProcessor",
]

import enum

from snmp.asn1 import *
from snmp.ber import *
from snmp.exception import *
from snmp.pdu import *
from snmp.smi import *
from snmp.typing import *
from snmp.utils import *

T = TypeVar("T")
TPDU = TypeVar("TPDU", bound=AnyPDU)
TMessage = TypeVar("TMessage", bound="Message")

@final
class BadVersion(IncomingMessageError):
    pass

class ProtocolVersion(enum.IntEnum):
    SNMPv1  = 0
    SNMPv2c = 1
    SNMPv3  = 3

    # Python 3.11 changes IntEnum.__str__()
    __str__ = enum.Enum.__str__

@final
class VersionOnlyMessage(Sequence):
    def __init__(self, version: ProtocolVersion) -> None:
        self.version = version

    def __iter__(self) -> Iterator[ASN1]:
        yield Integer(self.version)

    def __len__(self) -> int:
        return 1

    def __repr__(self) -> str:
        return f"{typename(self)}({str(self.version)})"

    @classmethod
    def deserialize(cls,
        data: Asn1Data,
    ) -> "VersionOnlyMessage":
        msgVersion, _ = cast(
            Tuple[Integer, subbytes],
            Integer.decode(data, leftovers=True),
        )

        try:
            version = ProtocolVersion(msgVersion.value)
        except ValueError as err:
            raise BadVersion(msgVersion.value) from err

        return cls(version)

class Message(Sequence):
    VERSIONS = (ProtocolVersion.SNMPv1, ProtocolVersion.SNMPv2c)

    def __init__(self,
        version: ProtocolVersion,
        community: bytes,
        pdu: AnyPDU,
    ) -> None:
        self.version = version
        self.community = community
        self.pdu = pdu

    def __iter__(self) -> Iterator[ASN1]:
        yield Integer(self.version)
        yield OctetString(self.community)
        yield self.pdu

    def __len__(self) -> int:
        return 3

    def __repr__(self) -> str:
        return "{}({}, {!r}, {})".format(
            typename(self),
            str(self.version),
            self.community,
            repr(self.pdu),
        )

    def __str__(self) -> str:
        return self.toString()

    def toString(self, depth: int = 0, tab: str = "    ") -> str:
        indent = tab * depth
        subindent = indent + tab
        return "\n".join((
            f"{indent}{typename(self)}:",
            f"{subindent}Community: {self.community!r}",
            f"{self.pdu.toString(depth+1, tab)}",
        ))

    @classmethod
    def deserialize(cls: Type[TMessage],
        data: Asn1Data,
        types: Optional[Mapping[Tag, Type[AnyPDU]]] = None,
    ) -> TMessage:
        msgVersion, ptr = cast(
            Tuple[Integer, subbytes],
            Integer.decode(data, leftovers=True),
        )

        try:
            version = ProtocolVersion(msgVersion.value)
        except ValueError as err:
            raise BadVersion(msgVersion.value) from err

        if version not in cls.VERSIONS:
            raise BadVersion(f"{typename} does not support {version.name}")

        community, ptr = cast(
            Tuple[OctetString, subbytes],
            OctetString.decode(ptr, leftovers=True),
        )

        identifier, _ = Tag.decode(subbytes(ptr))

        if types is None:
            types = dict()

        try:
            pduType = types[identifier]
        except KeyError as err:
            raise ParseError(f"Invalid PDU type: {identifier}") from err

        return cls(
            version,
            community.data,
            cast(AnyPDU, pduType.decode(ptr)),
        )

class RequestHandle(Generic[T]):
    def addCallback(self, func: Callable[[int], None], idNum: int) -> None:
        raise NotImplementedError()

    def push(self, response: T) -> None:
        raise NotImplementedError()

class MessageProcessor(Generic[T, TPDU]):
    VERSION: ClassVar[ProtocolVersion]

    def prepareDataElements(self, msg: Asn1Data) -> Tuple[T, RequestHandle[T]]:
        raise NotImplementedError()

    def prepareOutgoingMessage(self,
        pdu: TPDU,
        handle: RequestHandle[T],
        *args: Any,
        **kwargs: Any,
    ) -> bytes:
        raise NotImplementedError()



__all__ = ["AES_128_CFB128", "DES_CBC", "Decryptor", "Encryptor"]

from snmp.cffi import *
from snmp.openssl import *
from snmp.typing import *

class Cipher:
    def __init__(self,
        cipher: Callable[[], "Pointer[EVP_CIPHER]"],
        blocklen: int,
    ) -> None:
        self.BLOCKLEN = blocklen
        self.cipher = cipher

    def __call__(self) -> "Pointer[EVP_CIPHER]":
        return self.cipher()

AES_128_CFB128 = Cipher(lib.EVP_aes_128_cfb128, 16)
DES_CBC = Cipher(lib.EVP_des_cbc, 8)

class EnvelopeContext:
    def __init__(self) -> None:
        self.ctx = lib.EVP_CIPHER_CTX_new()

        if self.ctx == ffi.NULL:
            raise Exception("Failed to allocate EVP_CIPHER_CTX")

    def __enter__(self) -> "Pointer[EVP_CIPHER_CTX]":
        return self.ctx

    def __exit__(self, *args: Any) -> None:
        lib.EVP_CIPHER_CTX_free(self.ctx)

class Envelope:
    ENCRYPT: ClassVar[bool]

    def __init__(self, cipher: Cipher) -> None:
        self.cipher = cipher

        if self.ENCRYPT:
            self.init = lib.EVP_EncryptInit
            self.update = lib.EVP_EncryptUpdate
        else:
            self.init = lib.EVP_DecryptInit
            self.update = lib.EVP_DecryptUpdate

    @staticmethod
    def allocateOutputBuffer(
        datalen: int,
        blocklen: int,
    ) -> "UnsignedCharArray":
        length = datalen + blocklen - (datalen % blocklen)
        return ffi.new(f"unsigned char[]", length)

    def process(self, data: bytes, key: bytes, iv: bytes) -> bytes:
        with EnvelopeContext() as ctx:
            if not self.init(ctx, self.cipher(), key, iv):
                raise Exception("Failed to initialize cipher context")

            inl = len(data)
            outl = ffi.new("int*")
            output = self.allocateOutputBuffer(len(data), self.cipher.BLOCKLEN)

            if not self.update(ctx, output, outl, data, inl):
                raise Exception("Failed to update cipher context")

            return bytes(output)

class Decryptor(Envelope):
    ENCRYPT = False
    decrypt = Envelope.process

class Encryptor(Envelope):
    ENCRYPT = True
    encrypt = Envelope.process



__all__ = [
    "HeaderData", "MessageFlags", "ScopedPDU",
    "SNMPv3Message", "SNMPv3MessageProcessor",
]

import threading
import weakref

from snmp.asn1 import *
from snmp.ber import *
from snmp.exception import *
from snmp.message import *
from snmp.pdu import *
from snmp.security import *
from snmp.security.levels import *
from snmp.smi import *
from snmp.typing import *
from snmp.utils import *

pduTypes = {
    cls.TAG: cls for cls in cast(Tuple[Type[AnyPDU], ...], (
        GetRequestPDU,
        GetNextRequestPDU,
        ResponsePDU,
        SetRequestPDU,
        GetBulkRequestPDU,
        InformRequestPDU,
        SNMPv2TrapPDU,
        ReportPDU,
    ))
}

class CacheError(SNMPException):
    pass

class InvalidMessage(IncomingMessageError):
    pass

class LateResponse(IncomingMessageError):
    pass

class ResponseMismatch(IncomingMessageError):
    @classmethod
    def byField(cls, field: str) -> "ResponseMismatch":
        return cls(f"{field} does not match request")

class UnknownSecurityModel(IncomingMessageError):
    pass

@final
class MessageFlags(OctetString):
    AUTH_FLAG: ClassVar[int]        = (1 << 0)
    PRIV_FLAG: ClassVar[int]        = (1 << 1)
    REPORTABLE_FLAG: ClassVar[int]  = (1 << 2)

    def __init__(self,
        securityLevel: SecurityLevel = noAuthNoPriv,
        reportable: bool = False,
    ) -> None:
        self.securityLevel = securityLevel
        self.reportableFlag = reportable

    def __repr__(self) -> str:
        return f"{typename(self)}({self.securityLevel}, {self.reportableFlag})"

    def __str__(self) -> str:
        return self.toString()

    def toString(self, depth: int = 0, tab: str = "    ") -> str:
        indent = tab * depth
        subindent = indent + tab

        return "\n".join((
            f"{indent}{typename(self)}:",
            f"{subindent}Security Level: {self.securityLevel}",
            f"{subindent}Reportable: {self.reportableFlag}",
        ))

    @classmethod
    def construct(cls, data: Asn1Data = b"") -> "MessageFlags":
        try:
            byte = data[0]
        except IndexError as err:
            raise ParseError(f"{typename(cls)} must contain at least one byte")

        try:
            securityLevel = SecurityLevel(
                byte & cls.AUTH_FLAG,
                byte & cls.PRIV_FLAG,
            )
        except ValueError as err:
            raise InvalidMessage(f"Invalid msgFlags: {err}") from err

        reportable = (byte & cls.REPORTABLE_FLAG != 0)
        return cls(securityLevel, reportable)

    @property
    def data(self) -> bytes:
        byte = 0

        if self.authFlag:
            byte |= self.AUTH_FLAG

        if self.privFlag:
            byte |= self.PRIV_FLAG

        if self.reportableFlag:
            byte |= self.REPORTABLE_FLAG

        return bytes((byte,))

    @property
    def authFlag(self) -> bool:
        return self.securityLevel.auth

    @authFlag.setter
    def authFlag(self, value: Any) -> None:
        auth = bool(value)
        if auth != self.securityLevel.auth:
            self.securityLevel = SecurityLevel(
                auth,
                self.securityLevel.priv,
            )

    @property
    def privFlag(self) -> bool:
        return self.securityLevel.priv

    @privFlag.setter
    def privFlag(self, value: Any) -> None:
        priv = bool(value)
        if priv != self.securityLevel.priv:
            self.securityLevel = SecurityLevel(
                self.securityLevel.auth,
                priv,
            )

@final
class HeaderData(Sequence):
    def __init__(self,
        msgID: int,
        maxSize: int,
        flags: MessageFlags,
        securityModel: SecurityModel,
    ) -> None:
        self.id = msgID
        self.maxSize = maxSize
        self.flags = flags
        self.securityModel = securityModel

    def __iter__(self) -> Iterator[ASN1]:
        yield Integer(self.id)
        yield Integer(self.maxSize)
        yield self.flags
        yield Integer(self.securityModel)

    def __len__(self) -> int:
        return 4

    def __repr__(self) -> str:
        args = (
            str(self.id),
            str(self.maxSize),
            repr(self.flags),
            str(SecurityModel(self.securityModel)),
        )

        return f"{typename(self)}({', '.join(args)})"

    def __str__(self) -> str:
        return self.toString()

    def toString(self, depth: int = 0, tab: str = "    ") -> str:
        indent = tab * depth
        subindent = indent + tab
        securityModel = SecurityModel(self.securityModel)

        return "\n".join((
            f"{indent}{typename(self)}:",
            f"{subindent}Message ID: {self.id}",
            f"{subindent}Sender Message Size Limit: {self.maxSize}",
            f"{self.flags.toString(depth+1, tab)}",
            f"{subindent}Security Model: {securityModel.name}"
        ))

    @classmethod
    def deserialize(cls, data: Asn1Data) -> "HeaderData":
        msgID, data = cast(
            Tuple[Integer, subbytes],
            Integer.decode(data, leftovers=True),
        )

        msgMaxSize, data = cast(
            Tuple[Integer, subbytes],
            Integer.decode(data, leftovers=True),
        )

        msgFlags, data = cast(
            Tuple[MessageFlags, subbytes],
            MessageFlags.decode(data, leftovers=True),
        )

        msgSecurityModel = Integer.decode(data)

        if msgID.value < 0:
            raise ParseError("msgID may not be less than 0")
        elif msgMaxSize.value < 484:
            raise ParseError("msgMaxSize may not be less than 484")
        elif msgSecurityModel.value < 1:
            raise ParseError("msgSecurityModel may not be less than 1")

        try:
            securityModel = SecurityModel(msgSecurityModel.value)
        except ValueError as err:
            raise UnknownSecurityModel(msgSecurityModel.value) from err

        return cls(
            msgID.value,
            msgMaxSize.value,
            msgFlags,
            securityModel,
        )

@final
class ScopedPDU(Sequence):
    def __init__(self,
        pdu: AnyPDU,
        contextEngineID: bytes,
        contextName: bytes = b"",
    ) -> None:
        self.contextEngineID = contextEngineID
        self.contextName = contextName
        self.pdu = pdu

    def __iter__(self) -> Iterator[ASN1]:
        yield OctetString(self.contextEngineID)
        yield OctetString(self.contextName)
        yield self.pdu

    def __len__(self) -> int:
        return 3

    def __repr__(self) -> str:
        args = (
            repr(self.pdu),
            repr(self.contextEngineID),
            f"contextName={repr(self.contextName)}"
        )

        return f"{typename(self)}({', '.join(args)})"

    def __str__(self) -> str:
        return self.toString()

    def toString(self, depth: int = 0, tab: str = "    ") -> str:
        indent = tab * depth
        subindent = indent + tab
        return "\n".join((
            f"{indent}{typename(self)}:",
            f"{subindent}Context Engine ID: {self.contextEngineID!r}",
            f"{subindent}Context Name: {self.contextName!r}",
            f"{self.pdu.toString(depth=depth+1, tab=tab)}"
        ))

    @classmethod
    def deserialize(cls,
        data: Asn1Data,
        types: Optional[Mapping[Tag, Type[AnyPDU]]] = None,
    ) -> "ScopedPDU":
        if types is None:
            types = dict()

        contextEngineID, data = cast(
            Tuple[OctetString, subbytes],
            OctetString.decode(data, leftovers=True),
        )

        contextName, data = cast(
            Tuple[OctetString, subbytes],
            OctetString.decode(data, leftovers=True),
        )

        identifier, _ = Tag.decode(subbytes(data))

        try:
            pduType = types[identifier]
        except KeyError as err:
            raise ParseError(f"Invalid PDU type: {identifier}") from err

        return cls(
            cast(AnyPDU, pduType.decode(data)),
            contextEngineID = contextEngineID.data,
            contextName     = contextName.data,
        )

TMessage = TypeVar("TMessage", bound="SNMPv3Message")
class SNMPv3Message(Sequence):
    VERSION = ProtocolVersion.SNMPv3

    def __init__(self,
        header: HeaderData,
        scopedPDU: Optional[ScopedPDU] = None,
        encryptedPDU: Optional[OctetString] = None,
        securityParameters: Optional[OctetString] = None,
        securityEngineID: Optional[bytes] = None,
        securityName: Optional[bytes] = None,
    ) -> None:
        self.header = header
        self.scopedPDU = scopedPDU
        self.encryptedPDU = encryptedPDU

        if securityParameters is None:
            securityParameters = OctetString()

        self.securityParameters = securityParameters
        self.securityEngineID = securityEngineID
        self.securityName = securityName

    def __iter__(self) -> Iterator[ASN1]:
        yield Integer(self.VERSION)
        yield self.header
        yield self.securityParameters

        if self.header.flags.privFlag:
            if __debug__ and self.encryptedPDU is None:
                errmsg = "encryptedPDU is required when privFlag is True"
                raise SNMPLibraryBug(errmsg)

            yield self.encryptedPDU
        else:
            if __debug__ and self.scopedPDU is None:
                errmsg = "scopedPDU is required when privFlag is False"
                raise SNMPLibraryBug(errmsg)

            yield self.scopedPDU

    def __len__(self) -> int:
        return 4;

    def __repr__(self) -> str:
        args = [repr(self.header)]

        if self.header.flags.privFlag:
            args.append(f"encryptedPDU={repr(self.encryptedPDU)}")
        else:
            args.append(f"scopedPDU={repr(self.scopedPDU)}")

        args.append(f"securityParameters={repr(self.securityParameters)}")

        return f"{typename(self)}({', '.join(args)})"

    def __str__(self) -> str:
        return self.toString()

    def toString(self, depth: int = 0, tab: str = "    ") -> str:
        indent = tab * depth
        subindent = indent + tab

        if self.header.flags.privFlag:
            payload = f"{subindent}Encrypted Data: {self.encryptedPDU}"
        else:
            assert self.scopedPDU is not None
            payload = self.scopedPDU.toString(depth+1, tab)

        return "\n".join((
            f"{indent}{typename(self)}:",
            f"{self.header.toString(depth+1, tab)}",
            f"{subindent}Security Parameters: {self.securityParameters}",
            payload,
        ))

    @overload
    @classmethod
    def decode(
        cls: Type[TMessage],
        data: Asn1Data,
    ) -> TMessage:
        ...

    @overload
    @classmethod
    def decode(
        cls: Type[TMessage],
        data: Asn1Data,
        leftovers: bool = False,
        copy: bool = True,
        **kwargs: Any,
    ) -> Union[TMessage, Tuple[TMessage, subbytes]]:
        ...

    @classmethod
    def decode(cls: Type[TMessage],
        data: Asn1Data,
        leftovers: bool = False,
        copy: bool = False,
        **kwargs: Any,
    ) -> Union[TMessage, Tuple[TMessage, subbytes]]:
        return super().decode(data, leftovers, copy, **kwargs)

    @classmethod
    def deserialize(cls: Type[TMessage], data: Asn1Data) -> TMessage:
        msgVersion, ptr = cast(
            Tuple[Integer, subbytes],
            Integer.decode(data, leftovers=True),
        )

        try:
            version = ProtocolVersion(msgVersion.value)
        except ValueError as err:
            raise BadVersion(msgVersion.value) from err

        if version != cls.VERSION:
            raise BadVersion(f"{typename} does not support {version.name}")

        msgGlobalData, ptr = cast(
            Tuple[HeaderData, subbytes],
            HeaderData.decode(ptr, leftovers=True),
        )

        msgSecurityData, ptr = cast(
            Tuple[OctetString, subbytes],
            OctetString.decode(ptr, leftovers=True, copy=False),
        )

        scopedPDU = None
        encryptedPDU = None
        if msgGlobalData.flags.privFlag:
            encryptedPDU = OctetString.decode(ptr)
        else:
            scopedPDU = cast(ScopedPDU, ScopedPDU.decode(ptr, types=pduTypes))

        return cls(
            msgGlobalData,
            scopedPDU=scopedPDU,
            encryptedPDU=encryptedPDU,
            securityParameters=msgSecurityData,
        )

    @classmethod
    def findSecurityParameters(self, wholeMsg: bytes) -> subbytes:
        ptr: subbytes = decode(wholeMsg, self.TAG, copy=False)

        _, ptr = decode(ptr, Integer.TAG,       leftovers=True, copy=False)
        _, ptr = decode(ptr, Sequence.TAG,      leftovers=True, copy=False)
        ptr, _ = decode(ptr, OctetString.TAG,   leftovers=True, copy=False)

        return ptr

    @property
    def plaintext(self) -> bytes:
        assert self.scopedPDU is not None
        return self.scopedPDU.encode()

    @plaintext.setter
    def plaintext(self, data: bytes) -> None:
        self.scopedPDU, _ = cast(
            Tuple[ScopedPDU, subbytes],
            ScopedPDU.decode(data, leftovers=True, types=pduTypes),
        )

class CacheEntry:
    def __init__(self,
        engineID: bytes,
        contextName: bytes,
        handle: RequestHandle[SNMPv3Message],
        securityName: bytes,
        securityModel: SecurityModel,
        securityLevel: SecurityLevel,
    ):
        self.context = contextName
        self.engineID = engineID
        self.handle = weakref.ref(handle)
        self.securityName = securityName
        self.securityModel = securityModel
        self.securityLevel = securityLevel

class SNMPv3MessageProcessor(MessageProcessor[SNMPv3Message, AnyPDU]):
    VERSION = ProtocolVersion.SNMPv3

    def __init__(self, msgMaxSize: int) -> None:
        self.msgMaxSize = msgMaxSize

        self.cacheLock = threading.Lock()
        self.generator = self.newGenerator()
        self.outstanding: Dict[int, CacheEntry] = {}

        self.securityLock = threading.Lock()
        self.defaultSecurityModel: Optional[SecurityModel] = None
        self.securityModules: Dict[SecurityModel, SecurityModule[Any]] = {}

    @staticmethod
    def newGenerator() -> NumberGenerator:
        return NumberGenerator(31, signed=False)

    def cache(self, entry: CacheEntry) -> int:
        retry = 0
        while retry < 10:
            with self.cacheLock:
                msgID = next(self.generator)
                if msgID == 0:
                    self.generator = self.newGenerator()
                elif msgID not in self.outstanding:
                    self.outstanding[msgID] = entry
                    return msgID

            retry += 1

        raise CacheError("Failed to allocate message ID")

    def retrieve(self, msgID: int) -> CacheEntry:
        with self.cacheLock:
            return self.outstanding[msgID]

    def uncache(self, msgID: int) -> None:
        with self.cacheLock:
            try:
                del self.outstanding[msgID]
            except KeyError:
                pass

    def addSecurityModuleIfNeeded(self,
        module: SecurityModule[Any],
        default: bool = False,
    ) -> None:
        with self.securityLock:
            if module.MODEL not in self.securityModules:
                self.securityModules[module.MODEL] = module

                if default or self.defaultSecurityModel is None:
                    self.defaultSecurityModel = module.MODEL

    def prepareDataElements(self,
        msg: Asn1Data,
    ) -> Tuple[SNMPv3Message, RequestHandle[SNMPv3Message]]:
        message = SNMPv3Message.decode(msg)

        with self.securityLock:
            try:
                securityModule = self.securityModules[
                    message.header.securityModel
                ]
            except KeyError as e:
                securityModel = message.header.securityModel
                raise UnknownSecurityModel(securityModel) from e

        securityModule.processIncoming(message)

        if __debug__ and message.scopedPDU is None:
            errmsg = "securityModule.processIncoming() did not assign a" \
                "  value for message.scopedPDU"
            raise SNMPLibraryBug(errmsg)

        if isinstance(message.scopedPDU.pdu, Response):
            try:
                entry = self.retrieve(message.header.id)
            except KeyError as err:
                errmsg = f"Unknown msgID: {message.header.id}"
                raise ResponseMismatch(errmsg) from err

            handle = entry.handle()
            if handle is None:
                raise LateResponse("Handle has already been released")

            report = isinstance(message.scopedPDU.pdu, Internal)
            if (not report
            and message.header.flags.securityLevel < entry.securityLevel):
                raise ResponseMismatch.byField("Security Level")

            if not report and entry.engineID != message.securityEngineID:
                raise ResponseMismatch.byField("Security Engine ID")

            if entry.securityName != message.securityName:
                raise ResponseMismatch.byField("Security Name")

            if (not report
            and entry.engineID != message.scopedPDU.contextEngineID):
                raise ResponseMismatch.byField("Context Engine ID")

            if entry.context != message.scopedPDU.contextName:
                raise ResponseMismatch.byField("Context Name")
        else:
            raise UnsupportedFeature("Received a non-response PDU type")

        return message, handle

    def prepareOutgoingMessage(self,
        pdu: AnyPDU,
        handle: RequestHandle[SNMPv3Message],
        engineID: bytes,
        securityName: bytes,
        securityLevel: SecurityLevel = noAuthNoPriv,
        securityModel: Optional[SecurityModel] = None,
        contextName: bytes = b"",
    ) -> bytes:

        with self.securityLock:
            if securityModel is None:
                assert self.defaultSecurityModel is not None
                securityModel = self.defaultSecurityModel

            try:
                securityModule = self.securityModules[securityModel]
            except KeyError as err:
                errmsg = f"Security Model {securityModel} has not been enabled"
                raise ValueError(errmsg) from err

        reportable = isinstance(pdu, Confirmed)

        if reportable:
            entry = CacheEntry(
                engineID,
                contextName,
                handle,
                securityName,
                securityModel,
                securityLevel)

            msgID = self.cache(entry)
            handle.addCallback(self.uncache, msgID)
        else:
            # This feature is not yet implemented
            msgID = next(self.generator)

        flags = MessageFlags(securityLevel, reportable)
        header = HeaderData(msgID, self.msgMaxSize, flags, securityModel)
        scopedPDU = ScopedPDU(pdu, engineID, contextName=contextName)
        message = SNMPv3Message(header, scopedPDU)

        return securityModule.prepareOutgoing(message, engineID, securityName)



__all__ = ["UdpMultiplexor"]

import os
import select

from snmp.transport import *
from snmp.transport.udp import UdpSocket
from snmp.typing import *

class PosixUdpMultiplexor(TransportMultiplexor[Tuple[str, int]]):
    def __init__(self, recvSize: int = 1472) -> None:
        self.recvSize = recvSize
        self.r, self.w = os.pipe()
        self.sockets: Dict[int, UdpSocket] = {}

    def register(self, sock: UdpSocket) -> None:    # type: ignore[override]
        self.sockets[sock.fileno] = sock

    def listen(self, listener: TransportListener[Tuple[str, int]]) -> None:
        poller = select.poll()
        poller.register(self.r, select.POLLIN)

        for fileno in self.sockets.keys():
            poller.register(fileno, select.POLLIN)

        done = False
        while not done:
            ready = poller.poll()
            for fd, _ in ready:
                try:
                    sock = self.sockets[fd]
                except KeyError:
                    if fd == self.r:
                        os.read(fd, 1)
                        done = True
                else:
                    addr, data = sock.receive(self.recvSize)
                    listener.hear(sock, addr, data)

    def stop(self) -> None:
        os.write(self.w, bytes(1))

    def close(self) -> None:
        for sock in self.sockets.values():
            sock.close()

        os.close(self.w)
        os.close(self.r)

UdpMultiplexor = PosixUdpMultiplexor



__all__ = [
    "NoSuchObject", "NoSuchInstance", "EndOfMibView",
    "VarBind", "VarBindList",
    "AnyPDU", "PDU", "BulkPDU",
    "GetRequestPDU", "GetNextRequestPDU", "GetBulkRequestPDU",
    "SetRequestPDU",
    "ResponsePDU", "ReportPDU",
    "InformRequestPDU", "SNMPv2TrapPDU",
    "Read", "Write", "Response", "Internal", "Notification", "Confirmed",
    "ErrorStatus", "ErrorResponse",
]

import enum

from snmp.asn1 import *
from snmp.ber import *
from snmp.exception import *
from snmp.smi import *
from snmp.typing import *
from snmp.utils import subbytes, typename

AnyPDU = Union["PDU", "BulkPDU"]
TPDU = TypeVar("TPDU", bound="PDU")
TBulkPDU = TypeVar("TBulkPDU", bound="BulkPDU")

@final
class NoSuchObject(Null):
    TAG = Tag(0, cls = Tag.Class.CONTEXT_SPECIFIC)

@final
class NoSuchInstance(Null):
    TAG = Tag(1, cls = Tag.Class.CONTEXT_SPECIFIC)

@final
class EndOfMibView(Null):
    TAG = Tag(2, cls = Tag.Class.CONTEXT_SPECIFIC)

@final
class VarBind(Sequence):
    TYPES = {
        cls.TAG: cls for cls in cast(Tuple[Primitive, ...], (
            Integer,
            OctetString,
            Null,
            OID,
            IpAddress,
            Counter32,
            Gauge32,
            TimeTicks,
            Opaque,
            Counter64,
            NoSuchObject,
            NoSuchInstance,
            EndOfMibView,
        ))
    }

    def __init__(self,
        name: Union[str, OID],
        value: Optional[ASN1] = None,
    ) -> None:
        if not isinstance(name, OID):
            name = OID.parse(name)

        if value is None:
            value = Null()

        self.name = name
        self.value = value

    def __iter__(self) -> Iterator[ASN1]:
        yield self.name
        yield self.value

    def __len__(self) -> int:
        return 2

    def __repr__(self) -> str:
        return f"{typename(self)}({self.name!r}, {self.value!r})"

    def __str__(self) -> str:
        return f"{self.name}: {self.value}"

    @classmethod
    def deserialize(cls, data: Asn1Data) -> "VarBind":
        name, data = cast(
            Tuple[OID, subbytes],
            OID.decode(data, leftovers=True),
        )

        identifier, data = decode(data)

        try:
            valueType = cls.TYPES[identifier]
        except KeyError as err:
            msg = "Invalid variable value type: {}"
            raise ParseError(msg.format(identifier)) from err

        return cls(name, valueType.deserialize(data))

@final
class VarBindList(Sequence):
    def __init__(self, *args: Union[str, OID, VarBind]) -> None:
        self.variables = tuple(
            var if isinstance(var, VarBind) else VarBind(var) for var in args
        )

    def __bool__(self) -> bool:
        return bool(self.variables)

    @overload
    def __getitem__(self, key: int) -> VarBind:
        ...

    @overload
    def __getitem__(self, key: slice) -> Tuple[VarBind, ...]:
        ...

    def __getitem__(self,
        key: Union[int, slice],
    ) -> Union[VarBind, Tuple[VarBind, ...]]:
        return self.variables[key]

    def __iter__(self) -> Iterator[VarBind]:
        return iter(self.variables)

    def __len__(self) -> int:
        return len(self.variables)

    def __repr__(self) -> str:
        args = ", ".join(repr(var) for var in self.variables)
        return f"{typename(self)}({args})"

    def __str__(self) -> str:
        return self.toString()

    def toString(self, indent: str = "") -> str:
        return "\n".join(f"{indent}{var}" for var in self.variables)

    @classmethod
    def deserialize(cls, data: Asn1Data) -> "VarBindList":
        objects = []

        while data:
            var, data = cast(
                Tuple[VarBind, subbytes],
                VarBind.decode(data, leftovers=True),
            )

            objects.append(var)

        return cls(*objects)

class ErrorStatus(enum.IntEnum):
    noError             = 0
    tooBig              = 1
    noSuchName          = 2
    badValue            = 3
    readOnly            = 4
    genErr              = 5
    noAccess            = 6
    wrongType           = 7
    wrongLength         = 8
    wrongEncoding       = 9
    wrongValue          = 10
    noCreation          = 11
    inconsistentValue   = 12
    resourceUnavailable = 13
    commitFailed        = 14
    undoFailed          = 15
    authorizationError  = 16
    notWritable         = 17
    inconsistentName    = 18

class PDU(Constructed):
    def __init__(self,
        *args: Union[str, OID, VarBind],
        requestID: int = 0,
        errorStatus: ErrorStatus = ErrorStatus.noError,
        errorIndex: int = 0,
        variableBindings: Optional[VarBindList] = None,
    ) -> None:
        self.requestID = requestID
        self.errorStatus = errorStatus
        self.errorIndex = errorIndex

        if variableBindings is None:
            self.variableBindings = VarBindList(*args)
        else:
            self.variableBindings = variableBindings

    def __iter__(self) -> Iterator[ASN1]:
        yield Integer(self.requestID)
        yield Integer(self.errorStatus)
        yield Integer(self.errorIndex)
        yield self.variableBindings

    def __len__(self) -> int:
        return 4

    def __repr__(self) -> str:
        arguments = []
        if self.requestID:
            arguments.append("requestID={}".format(self.requestID))

        if self.errorStatus:
            arguments.append("errorStatus={}".format(self.errorStatus))

        if self.errorIndex:
            arguments.append("errorIndex={}".format(self.errorIndex))

        arguments.append(f"variableBindings={self.variableBindings!r}")

        args = ", ".join(arguments)
        return f"{typename(self)}({args})"

    def __str__(self) -> str:
        return self.toString()

    def toString(self, depth: int = 0, tab: str = "    ") -> str:
        indent = tab * depth
        subindent = indent + tab
        return "\n".join((
            "{}{}:",
            "{}Request ID: {}",
            "{}Error Status: {}",
            "{}Error Index: {}",
            "{}Variable Bindings:",
            "{}"
        )).format(
            indent, typename(self),
            subindent, self.requestID,
            subindent, self.errorStatus.name,
            subindent, self.errorIndex,
            subindent, self.variableBindings.toString(subindent + tab)
        )

    @classmethod
    def deserialize(cls: Type[TPDU], data: Asn1Data) -> TPDU:
        _requestID, data = cast(
            Tuple[Integer, subbytes],
            Integer.decode(data, leftovers=True),
        )

        _errorStatus, data = cast(
            Tuple[Integer, subbytes],
            Integer.decode(data, leftovers=True),
        )

        _errorIndex, data = cast(
            Tuple[Integer, subbytes],
            Integer.decode(data, leftovers=True),
        )

        variableBindings = VarBindList.decode(data)

        requestID = _requestID.value
        errorStatus = _errorStatus.value
        errorIndex = _errorIndex.value

        try:
            errorStatus = ErrorStatus(errorStatus)
        except ValueError as err:
            msg = "Invalid errorStatus: {}"
            raise ParseError(msg.format(errorStatus)) from err

        if (errorStatus != ErrorStatus.noError
        and errorIndex < 0 or errorIndex > len(variableBindings)):
            msg = "Error index {} not valid with {} variable bindings"
            raise ParseError(msg.format(errorIndex, len(variableBindings)))

        return cls(
            requestID=requestID,
            errorStatus=errorStatus,
            errorIndex=errorIndex,
            variableBindings=variableBindings,
        )

class BulkPDU(Constructed):
    def __init__(self,
        *args: Union[str, OID, VarBind],
        requestID: int = 0,
        nonRepeaters: int = 0,
        maxRepetitions: int = 0,
        variableBindings: Optional[VarBindList] = None,
    ) -> None:
        self.requestID = requestID
        self.nonRepeaters = nonRepeaters
        self.maxRepetitions = maxRepetitions

        if variableBindings is None:
            self.variableBindings = VarBindList(*args)
        else:
            self.variableBindings = variableBindings

    def __iter__(self) -> Iterator[ASN1]:
        yield Integer(self.requestID)
        yield Integer(self.nonRepeaters)
        yield Integer(self.maxRepetitions)
        yield self.variableBindings

    def __len__(self) -> int:
        return 4

    def __repr__(self) -> str:
        arguments = []
        if self.requestID:
            arguments.append("requestID={}".format(self.requestID))

        if self.nonRepeaters:
            arguments.append("nonRepeaters={}".format(self.nonRepeaters))

        if self.maxRepetitions:
            arguments.append("maxRepetitions={}".format(self.maxRepetitions))

        arguments.append("variableBindings={}".format(repr(self.variableBindings)))

        args = ", ".join(arguments)
        return f"{typename(self)}({args})"

    def __str__(self) -> str:
        return self.toString()

    def toString(self, depth: int = 0, tab: str = "    ") -> str:
        indent = tab * (depth + 1)
        return "\n".join((
            "{}:",
            "{}Request ID: {}",
            "{}Non-Repeaters: {}",
            "{}Max Repetitions: {}",
            "{}Variable Bindings:",
            "{}"
        )).format(
            typename(self),
            indent, self.requestID,
            indent, self.nonRepeaters,
            indent, self.maxRepetitions,
            indent, self.variableBindings.toString(tab * (depth + 2))
        )

    @classmethod
    def deserialize(cls: Type[TBulkPDU], data: Asn1Data) -> TBulkPDU:
        requestID, data = cast(
            Tuple[Integer, subbytes],
            Integer.decode(data, leftovers=True),
        )

        nonRepeaters, data = cast(
            Tuple[Integer, subbytes],
            Integer.decode(data, leftovers=True),
        )

        maxRepetitions, data = cast(
            Tuple[Integer, subbytes],
            Integer.decode(data, leftovers=True),
        )

        variableBindings = VarBindList.decode(data)

        if nonRepeaters.value < 0:
            raise ParseError("nonRepeaters may not be less than 0")
        elif maxRepetitions.value < 0:
            raise ParseError("maxRepetitions may not be less than 0")

        return cls(
            requestID=requestID.value,
            nonRepeaters=nonRepeaters.value,
            maxRepetitions=maxRepetitions.value,
            variableBindings=variableBindings,
        )

class Read:
    pass

class Write:
    pass

class Response:
    pass

class Notification:
    pass

class Internal:
    pass

class Confirmed:
    pass

@final
class GetRequestPDU(PDU, Read, Confirmed):
    TAG = Tag(0, True, Tag.Class.CONTEXT_SPECIFIC)

@final
class GetNextRequestPDU(PDU, Read, Confirmed):
    TAG = Tag(1, True, Tag.Class.CONTEXT_SPECIFIC)

@final
class ResponsePDU(PDU, Response):
    TAG = Tag(2, True, Tag.Class.CONTEXT_SPECIFIC)

@final
class SetRequestPDU(PDU, Write, Confirmed):
    TAG = Tag(3, True, Tag.Class.CONTEXT_SPECIFIC)

@final
class GetBulkRequestPDU(BulkPDU, Read, Confirmed):
    TAG = Tag(5, True, Tag.Class.CONTEXT_SPECIFIC)

@final
class InformRequestPDU(PDU, Notification, Confirmed):
    TAG = Tag(6, True, Tag.Class.CONTEXT_SPECIFIC)

@final
class SNMPv2TrapPDU(PDU, Notification):
    TAG = Tag(7, True, Tag.Class.CONTEXT_SPECIFIC)

@final
class ReportPDU(PDU, Response, Internal):
    TAG = Tag(8, True, Tag.Class.CONTEXT_SPECIFIC)

class ErrorResponse(SNMPException):
    def __init__(self,
        status: ErrorStatus,
        index: int,
        request: "AnyPDU",
    ) -> None:
        self.status = status
        self.cause: Union[VarBind, AnyPDU, int] = None

        details = ""
        if index == 0:
            self.cause = request
        else:
            try:
                self.cause = request.variableBindings[index-1]
            except IndexError:
                self.cause = index
            else:
                details = f": {self.cause.name}"

        super().__init__(f"{status.name}{details}")



__all__ = ["AesCfb128"]

import os

from snmp.security.usm import DecryptionError, PrivProtocol
from snmp.typing import *

from . import *

class AesCfb128(PrivProtocol):
    BYTEORDER:  ClassVar[Literal["big"]] = "big"
    CIPHER = AES_128_CFB128

    INTSIZE:    ClassVar[int] = 4
    KEYLEN:     ClassVar[int] = CIPHER.BLOCKLEN
    SALTLEN:    ClassVar[int] = CIPHER.BLOCKLEN - (2 * INTSIZE)
    SALTWRAP:   ClassVar[int] = 1 << (8 * SALTLEN)

    def __init__(self, key: bytes) -> None:
        if len(key) < self.KEYLEN:
            errmsg = "key must be at least {} bytes long".format(self.KEYLEN)
            raise ValueError(errmsg)

        self.key = key[:self.KEYLEN]
        self.salt = int.from_bytes(os.urandom(self.SALTLEN), self.BYTEORDER)

    def packIV(self, engineBoots: int, engineTime: int, salt: bytes) -> bytes:
        if len(salt) != self.SALTLEN:
            raise ValueError("Invalid salt")

        return b''.join((
            engineBoots.to_bytes(self.INTSIZE, self.BYTEORDER),
            engineTime .to_bytes(self.INTSIZE, self.BYTEORDER),
            salt
        ))

    def decrypt(self,
        data: bytes,
        engineBoots: int,
        engineTime: int,
        salt: bytes,
    ) -> bytes:
        try:
            iv = self.packIV(engineBoots, engineTime, salt)
        except ValueError as err:
            raise DecryptionError(err) from err

        return Decryptor(self.CIPHER).decrypt(data, self.key, iv)

    def encrypt(self,
        data: bytes,
        engineBoots: int,
        engineTime: int,
    ) -> Tuple[bytes, bytes]:
        self.salt = (self.salt + 1) % self.SALTWRAP

        salt = self.salt.to_bytes(self.SALTLEN, self.BYTEORDER)
        iv = self.packIV(engineBoots, engineTime, salt)
        ciphertext = Encryptor(self.CIPHER).encrypt(data, self.key, iv)

        return ciphertext, salt



__all__ = ["FFI", "Pointer", "UnsignedCharArray"]

from snmp.typing import *

T = TypeVar("T")
class Pointer(Generic[T]):
    def __getitem__(self, index: Literal[0]) -> T: ...

class UnsignedCharArray:
    def __bytes__(self) -> bytes: ...

class FFI:
    def memmove(self, dest: Pointer[T], src: Pointer[T], n: int) -> None: ...
    def sizeof(self, typename: str) -> int: ...



__all__ = [
    "ASN1", "Constructed", "Primitive",
    "INTEGER", "OCTET_STRING", "NULL", "OBJECT_IDENTIFIER", "SEQUENCE",
]

from snmp.ber import *
from snmp.exception import *
from snmp.typing import *
from snmp.utils import *

TASN1           = TypeVar("TASN1",          bound="ASN1")
TPrimitive      = TypeVar("TPrimitive",     bound="Primitive")
TINTEGER        = TypeVar("TINTEGER",       bound="INTEGER")
TOCTET_STRING   = TypeVar("TOCTET_STRING",  bound="OCTET_STRING")
TNULL           = TypeVar("TNULL",          bound="NULL")
TOID            = TypeVar("TOID",           bound="OBJECT_IDENTIFIER")

class ASN1:
    TAG: ClassVar[Tag]

    @overload
    @classmethod
    def decode(
        cls: Type[TASN1],
        data: Asn1Data,
    ) -> TASN1:
        ...

    @overload
    @classmethod
    def decode(
        cls: Type[TASN1],
        data: Asn1Data,
        leftovers: bool = False,
        copy: bool = True,
        **kwargs: Any,
    ) -> Union[TASN1, Tuple[TASN1, subbytes]]:
        ...

    @classmethod
    def decode(
        cls: Type[TASN1],
        data: Asn1Data,
        leftovers: bool = False,
        copy: bool = True,
        **kwargs: Any,
    ) -> Union[TASN1, Tuple[TASN1, subbytes]]:
        if leftovers:
            encoding, tail = decode(data, cls.TAG, True, copy)
            return cls.deserialize(encoding, **kwargs), tail
        else:
            encoding = decode(data, cls.TAG, False, copy)
            return cls.deserialize(encoding, **kwargs)

    def encode(self) -> bytes:
        return encode(self.TAG, self.serialize())

    @classmethod
    def deserialize(cls: Type[TASN1], data: Asn1Data) -> TASN1:
        raise NotImplementedError()

    def serialize(self) -> bytes:
        raise NotImplementedError()

class Constructed(ASN1):
    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Constructed):
            return NotImplemented

        if type(self) != type(other) or len(self) != len(other):
            return False

        for left, right in zip(self, other):
            if left != right:
                return False

        return True

    def __iter__(self) -> Iterator[ASN1]:
        raise NotImplementedError()

    def __len__(self) -> int:
        raise NotImplementedError()

    def serialize(self) -> bytes:
        return b"".join([obj.encode() for obj in self])

class Primitive(ASN1):
    def asOID(self, implied: bool = False) -> Iterable[int]:
        raise NotImplementedError()

    @classmethod
    def fromOID(
        cls: Type[TPrimitive],
        nums: Iterator[int],
        implied: bool = False,
    ) -> TPrimitive:
        raise NotImplementedError()

class INTEGER(Primitive):
    BYTEORDER: ClassVar[Literal["big"]] = "big"
    TAG = Tag(2)

    def __init__(self, value: int) -> None:
        self._value = value

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, INTEGER):
            return NotImplemented

        return self.value == other.value and self.TAG == other.TAG

    def __repr__(self) -> str:
        return f"{typename(self)}({self.value})"

    @property
    def value(self) -> int:
        return self._value

    @staticmethod
    def bitCount(value: int) -> int:
        if value < 0:
            value = -value - 1

        return value.bit_length()

    @classmethod
    def construct(cls: Type[TINTEGER], value: int) -> TINTEGER:
        return cls(value)

    @classmethod
    def deserialize(cls: Type[TINTEGER], data: Asn1Data) -> TINTEGER:
        value = int.from_bytes(data, cls.BYTEORDER, signed=True)
        return cls.construct(value)

    def serialize(self) -> bytes:
        # equivalent to (N + 8) // 8
        # the reason it's not (N + 7) is that ASN.1 always includes a sign bit
        nbytes = (self.bitCount(self.value) // 8) + 1
        return self.value.to_bytes(nbytes, self.BYTEORDER, signed=True)

    def asOID(self, implied: bool = False) -> Iterable[int]:
        yield self.value

    @classmethod
    def fromOID(
        cls: Type[TINTEGER],
        nums: Iterator[int],
        implied: bool = False,
    ) -> TINTEGER:
        try:
            return cls.construct(next(nums))
        except ParseError as err:
            raise OBJECT_IDENTIFIER.IndexDecodeError(*err.args) from err

class OCTET_STRING(Primitive):
    TAG = Tag(4)

    def __init__(self, data: bytes = b"") -> None:
        self._data = data

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, OCTET_STRING):
            return NotImplemented

        return self.data == other.data and self.TAG == other.TAG

    def __repr__(self) -> str:
        return f"{typename(self)}({repr(self.data)})"

    @property
    def data(self) -> bytes:
        return self._data

    def asOID(self, implied: bool = False) -> Iterable[int]:
        if not implied:
            yield len(self.data)

        yield from self.data

    @classmethod
    def fromOID(
        cls: Type[TOCTET_STRING],
        nums: Iterator[int],
        implied: bool = False,
    ) -> TOCTET_STRING:
        if implied:
            data = bytes(nums)
        else:
            length = next(nums)
            data = bytes([next(nums) for _ in range(length)])

        try:
            return cls.construct(data)
        except ParseError as err:
            raise OBJECT_IDENTIFIER.IndexDecodeError(*err.args) from err

    @classmethod
    def construct(cls: Type[TOCTET_STRING], data: Asn1Data) -> TOCTET_STRING:
        if isinstance(data, subbytes):
            data = data[:]

        return cls(data)

    @classmethod
    def deserialize(cls: Type[TOCTET_STRING], data: Asn1Data) -> TOCTET_STRING:
        return cls.construct(data)

    def serialize(self) -> bytes:
        return self.data

class NULL(Primitive):
    TAG = Tag(5)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, NULL):
            return NotImplemented

        return self.TAG == other.TAG

    def __repr__(self) -> str:
        return f"{typename(self)}()"

    def asOID(self, implied: bool = False) -> Iterable[int]:
        return ()

    @classmethod
    def fromOID(
        cls: Type[TNULL],
        nums: Iterator[int],
        implied: bool = False,
    ) -> TNULL:
        return cls()

    @classmethod
    def deserialize(cls: Type[TNULL], data: Asn1Data) -> TNULL:
        return cls()

    def serialize(self) -> bytes:
        return b""

class OBJECT_IDENTIFIER(Primitive):
    TAG = Tag(6)

    class BadPrefix(SNMPException):
        pass

    class IndexDecodeError(SNMPException):
        pass

    def __init__(self, *subidentifiers: int) -> None:
        first = 0
        second = 0

        try:
            first = subidentifiers[0]
            second = subidentifiers[1]
        except IndexError:
            pass

        if first > 2:
            raise ValueError("The first subidentifier must be less than 3")
        elif second >= 40:
            raise ValueError("The second subidentifier must be less than 40")

        if any(map(lambda x: x < 0, subidentifiers)):
            raise ValueError("Sub-identifiers may not be negative")

        self.subidentifiers = subidentifiers

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, OBJECT_IDENTIFIER):
            return NotImplemented

        return self.subidentifiers == other.subidentifiers

    def __lt__(self, other: "OBJECT_IDENTIFIER") -> bool:
        return self.subidentifiers < other.subidentifiers

    @overload
    def __getitem__(self, index: int) -> int:
        ...

    @overload
    def __getitem__(self, index: slice) -> Tuple[int, ...]:
        ...

    def __getitem__(self,
        index: Union[int, slice],
    ) -> Union[int, Tuple[int, ...]]:
        return self.subidentifiers.__getitem__(index)

    def __hash__(self) -> int:
        return self.subidentifiers.__hash__()

    def __iter__(self) -> Iterator[int]:
        return iter(self.subidentifiers)

    def __len__(self) -> int:
        return len(self.subidentifiers)

    def __repr__(self) -> str:
        args = ", ".join(str(i) for i in self.subidentifiers)
        return f"{typename(self)}({args})"

    def __str__(self) -> str:
        return ".".join(str(i) for i in  self.subidentifiers)

    @classmethod
    def parse(cls: Type[TOID], oid: str) -> TOID:
        numbers = oid.split(".")

        if numbers[0] == "":
            numbers = numbers[1:]

        try:
            return cls(*(int(n) for n in numbers))
        except ValueError as err:
            raise ValueError(f"Invalid {typename(cls)} string: {oid}") from err

    def asOID(self, implied: bool = False) -> Iterable[int]:
        if not implied:
            yield len(self.subidentifiers)

        yield from self.subidentifiers

    @classmethod
    def fromOID(
        cls: Type[TOID],
        nums: Iterator[int],
        implied: bool = False,
    ) -> TOID:
        if implied:
            subidentifiers = nums
        else:
            length = next(nums)
            subidentifiers = (next(nums) for _ in range(length))

        try:
            return cls.construct(*subidentifiers)
        except ParseError as err:
            raise OBJECT_IDENTIFIER.IndexDecodeError(*err.args) from err

    def extend(self: TOID, *subidentifiers: int) -> TOID:
        return self.__class__(*self.subidentifiers, *subidentifiers)

    def startswith(self, prefix: "OBJECT_IDENTIFIER") -> bool:
        return self.subidentifiers[:len(prefix)] == prefix[:]

    def decodeIndex(self: TOID,
        prefix: TOID,
        *types: Type[TPrimitive],
        implied: bool = False,
    ) -> Tuple[TPrimitive, ...]:
        if not self.startswith(prefix):
            raise self.BadPrefix(f"{self} does not begin with {prefix}")

        nums = iter(self.subidentifiers[len(prefix):])

        index = []
        for cls in types[:-1]:
            index.append(self.decodeIndexField(nums, cls))

        for cls in types[-1:]:
            index.append(self.decodeIndexField(nums, cls, implied=implied))

        try:
            next(nums)
        except StopIteration:
            pass
        else:
            errmsg = "Not all sub-identifiers were consumed"
            raise self.IndexDecodeError(errmsg)

        return tuple(index)

    @staticmethod
    def decodeIndexField(
        nums: Iterator[int],
        cls: Type[TPrimitive],
        implied: bool = False,
    ) -> TPrimitive:
        try:
            return cls.fromOID(nums, implied=implied)
        except StopIteration as err:
            errmsg = f"Incomplete {typename(cls)} field in index"
            raise OBJECT_IDENTIFIER.IndexDecodeError(errmsg)

    def withIndex(self: TOID,
        *index: Primitive,
        implied: bool = False,
    ) -> TOID:
        oid = self

        for obj in index[:-1]:
            oid = oid.extend(*obj.asOID())

        for obj in index[-1:]:
            oid = oid.extend(*obj.asOID(implied=implied))

        return oid

    @classmethod
    def construct(cls: Type[TOID], *subidentifiers: int) -> TOID:
        try:
            return cls(*subidentifiers)
        except ValueError as err:
            raise ParseError(*err.args) from err

    @classmethod
    def deserialize(cls: Type[TOID], data: Asn1Data) -> TOID:
        stream = iter(data)

        try:
            oid = list(divmod(next(stream), 40))
        except StopIteration as err:
            raise ParseError(f"Empty {typename(cls)}") from err

        value = 0
        for byte in stream:
            value |= byte & 0x7f
            if byte & 0x80:
                value <<= 7
            else:
                oid.append(value)
                value = 0

        if value:
            raise ParseError(f"{typename(cls)} ended unexpectedly")

        return cls.construct(*oid)

    @staticmethod
    def serializeSubIdentifier(bytearr: bytearray, num: int) -> None:
        if num < 0x80:
            bytearr.append(num)
        else:
            flag = 0
            tmp = bytearray()

            while num:
                tmp.append((num & 0x7f) | flag)
                flag = 0x80
                num >>= 7

            tmp.reverse()
            bytearr.extend(tmp)

    def serialize(self) -> bytes:
        try:
            first = self.subidentifiers[0]
        except IndexError:
            return b"\x00"

        try:
            second = self.subidentifiers[1]
        except IndexError:
            second = 0

        encoding = bytearray()
        self.serializeSubIdentifier(encoding, first * 40 | second)
        for subidentifier in self.subidentifiers[2:]:
            self.serializeSubIdentifier(encoding, subidentifier)

        return bytes(encoding)

class SEQUENCE(Constructed):
    TAG = Tag(16, True)



__all__ = ["SNMPv1Manager"]

import heapq
import math
import threading
import time

from snmp.asn1 import *
from snmp.dispatcher import *
from snmp.manager import *
from snmp.message import *
from snmp.pdu import *
from snmp.smi import *
from snmp.transport import *
from snmp.typing import *
from snmp.utils import *

class Request(RequestHandle[Message]):
    def __init__(self,
        pdu: PDU,
        manager: "SNMPv1Manager[Any]",
        community: bytes,
        timeout: float = 10.0,
        refreshPeriod: float = 1.0,
    ) -> None:
        now = time.time()

        self.community = community
        self.manager = manager
        self.pdu = pdu

        self.callback: Optional[Callable[[int], None]] = None
        self.event = threading.Event()
        self.response: Optional[Message] = None

        self.expiration = now + timeout
        self._nextRefresh = self.expiration
        self.period = refreshPeriod

    def __del__(self) -> None:
        if self.callback is not None:
            self.close()

    @property
    def expired(self) -> bool:
        return self.expiration <= time.time()

    @property
    def nextRefresh(self) -> float:
        return self._nextRefresh

    @nextRefresh.setter
    def nextRefresh(self, value: float) -> None:
        self._nextRefresh = min(self.expiration, value)

    def close(self) -> None:
        assert self.callback is not None
        self.callback(self.pdu.requestID)
        self.callback = None

    def addCallback(self,
        callback: Callable[[int], None],
        requestID: int,
    ) -> None:
        assert requestID == self.pdu.requestID
        assert self.callback is None

        self.callback = callback

    def push(self, response: Message) -> None:
        self.response = response
        self.event.set()

    # Always update self.nextRefresh right before calling this method
    def reallySend(self) -> None:
        self.manager.sendPdu(self.pdu, self, self.community)

    def refresh(self) -> Optional[float]:
        if self.event.is_set():
            return None

        now = time.time()
        timeToNextRefresh = self.nextRefresh - now

        if timeToNextRefresh <= 0.0:
            if self.expiration <= now:
                return None

            # Calculating it like this mitigates over-delay
            periodsElapsed = math.ceil(-timeToNextRefresh / self.period)
            self.nextRefresh += periodsElapsed * self.period
            self.reallySend()
            return 0.0
        else:
            return timeToNextRefresh

    def send(self) -> None:
        now = time.time()
        self.nextRefresh = now + self.period
        self.reallySend()

    def wait(self) -> VarBindList:
        pdu: Optional[ResponsePDU] = None
        while not self.expired:
            timeout = self.manager.refresh()
            if self.event.wait(timeout=timeout):
                assert self.response is not None
                pdu = cast(ResponsePDU, self.response.pdu)
                break

        self.close()
        if pdu is None:
            raise Timeout()
        else:
            if pdu.errorStatus:
                raise ErrorResponse(pdu.errorStatus, pdu.errorIndex, self.pdu)
            else:
                return pdu.variableBindings

T = TypeVar("T")
class SNMPv1Manager(Generic[T]):
    def __init__(self,
        dispatcher: Dispatcher[T],
        channel: TransportChannel[T],
        community: bytes,
        autowait: bool = True,
    ):
        self.autowait = autowait
        self.channel = channel

        self.dispatcher = dispatcher
        self.defaultCommunity = community

        self.lock = threading.Lock()
        self.requests: List[ComparableWeakRef[Request, float]] = []

    def refresh(self) -> Optional[float]:
        while self.requests:
            with self.lock:
                reference = self.requests[0]
                request = reference()

                if request is None:
                    wait = None
                else:
                    wait = request.refresh()

                if wait is None:
                    heapq.heappop(self.requests)
                    continue
                elif wait > 0.0:
                    return wait
                else:
                    heapq.heapreplace(self.requests, reference)

        return None

    def sendPdu(self,
        pdu: PDU,
        handle: Request,
        community: bytes,
    ) -> None:
        self.dispatcher.sendPdu(
            self.channel,
            ProtocolVersion.SNMPv1,
            pdu,
            handle,
            community,
        )

    def sendRequest(self,
        pdu: PDU,
        community: Optional[bytes] = None,
        wait: Optional[bool] = None,
        **kwargs: Any,
    ) -> Union[Request, VarBindList]:
        if community is None:
            community = self.defaultCommunity

        if wait is None:
            wait = self.autowait

        request = Request(pdu, self, community, **kwargs)
        reference = ComparableWeakRef(request, key=lambda r: r.nextRefresh)

        with self.lock:
            heapq.heappush(self.requests, reference)
            request.send()

        if wait:
            return request.wait()
        else:
            return request

    def get(self,
        *oids: Union[str, OID],
        **kwargs: Any,
    ) -> Union[Request, VarBindList]:
        pdu = GetRequestPDU(*oids)
        return self.sendRequest(pdu, **kwargs)

    def getNext(self,
        *oids: Union[str, OID],
        **kwargs: Any,
    ) -> Union[Request, VarBindList]:
        pdu = GetNextRequestPDU(*oids)
        return self.sendRequest(pdu, **kwargs)

    def set(self,
        *varbinds: Union[
            Tuple[
                Union[str, OID],
                Optional[ASN1],
            ],
            VarBind,
        ],
        **kwargs: Any,
    ) -> Union[Request, VarBindList]:
        vbList = (VarBind(*varbind) for varbind in varbinds)
        pdu = SetRequestPDU(*vbList)
        return self.sendRequest(pdu, **kwargs)



__all__ = ["SNMPv3UsmManager"]

from collections import deque

import heapq
import math
import threading
import time
import weakref

from snmp.asn1 import *
from snmp.dispatcher import *
from snmp.exception import *
from snmp.message import *
from snmp.message.v3 import *
from snmp.pdu import *
from snmp.security import *
from snmp.security.levels import *
from snmp.security.usm import *
from snmp.smi import *
from snmp.transport import *
from snmp.typing import *
from snmp.utils import *
from . import *

usmStats = OID.parse("1.3.6.1.6.3.15.1.1")
usmStatsUnsupportedSecLevels        = usmStats.extend(1, 0)
usmStatsNotInTimeWindowsInstance    = usmStats.extend(2, 0)
usmStatsUnknownUserNames            = usmStats.extend(3, 0)
usmStatsUnknownEngineIDsInstance    = usmStats.extend(4, 0)
usmStatsWrongDigests                = usmStats.extend(5, 0)
usmStatsDecryptionErrors            = usmStats.extend(6, 0)

class UnhandledReport(SNMPException):
    pass

class UnrecognizedReport(UnhandledReport):
    def __init__(self, varbind: VarBind) -> None:
        super().__init__(f"Remote Engine sent this Report: \"{varbind}\"")
        self.name = varbind.name
        self.value = varbind.value

class UsmStatsReport(UnhandledReport):
    def __init__(self, errmsg: Any, varbind: VarBind) -> None:
        super().__init__(errmsg)
        self.name = varbind.name
        self.value = varbind.value

class UnsupportedSecurityLevelReport(UsmStatsReport):
    def __init__(self, message: SNMPv3Message, varbind: VarBind) -> None:
        securityLevel = message.header.flags.securityLevel
        errmsg = f"Remote Engine does not support {securityLevel}"
        super().__init__(errmsg, varbind)

class UnknownUserNameReport(UsmStatsReport):
    def __init__(self, message: SNMPv3Message, varbind: VarBind) -> None:
        assert message.securityName is not None
        userName = message.securityName.decode()
        errmsg = f"Remote Engine does not recognize user \"{userName}\""
        super().__init__(errmsg, varbind)

class WrongDigestReport(UsmStatsReport):
    def __init__(self, message: SNMPv3Message, varbind: VarBind) -> None:
        assert message.securityName is not None
        userName = message.securityName.decode()

        errmsg = (
            "Remote Engine reported that your request failed authentication"
            f"; make sure \"{userName}\" is using the correct credentials"
        )

        super().__init__(errmsg, varbind)

class DecryptionErrorReport(UsmStatsReport):
    def __init__(self, message: SNMPv3Message, varbind: VarBind) -> None:
        errmsg = "Remote Engine was unable to decrypt your request"
        super().__init__(errmsg, varbind)

class DiscoveryError(IncomingMessageError):
    pass

class State:
    def __init__(self, arg: Union["SNMPv3UsmManager[Any]", "State"]):
        self.manager: "SNMPv3UsmManager[Any]"
        if isinstance(arg, State):
            self.manager = arg.manager
        else:
            self.manager = weakref.proxy(arg)

    def onInactive(self) -> None:
        raise NotImplementedError()

    def onRequest(self, auth: bool) -> bool:
        raise NotImplementedError()

    def onReport(self, engineID: bytes) -> bool:
        raise NotImplementedError()

    def onResponse(self, engineID: bytes, auth: bool) -> bool:
        raise NotImplementedError()

# User did not provide an engine ID, and
# has not yet attempted to send any requests
class Inactive(State):
    def onInactive(self) -> None:
        pass

    def onRequest(self, auth: bool) -> bool:
        self.manager.nextState = WaitForDiscovery(self)
        return True

    def onReport(self, engineID: bytes) -> bool:
        errmsg = "Received a report where no request should have been sent"
        raise SNMPLibraryBug(errmsg)

    def onResponse(self, engineID: bytes, auth: bool) -> bool:
        errmsg = "Received a response where no request should have been sent"
        raise SNMPLibraryBug(errmsg)

# User provided an engine ID, but has not yet sent any requests
class Unsynchronized(State):
    def onInactive(self) -> None:
        pass

    def onRequest(self, auth: bool) -> bool:
        self.manager.nextState = Synchronizing(self)
        return True

    def onReport(self, engineID: bytes) -> bool:
        errmsg = "Received a report where no request should have been sent"
        raise SNMPLibraryBug(errmsg)

    def onResponse(self, engineID: bytes, auth: bool) -> bool:
        errmsg = "Received a response where no request should have been sent"
        raise SNMPLibraryBug(errmsg)

# User did not provide an engine ID, and the Manager has not
# yet received any communication from the remote engine.
class WaitForDiscovery(State):
    def onInactive(self) -> None:
        self.manager.nextState = Inactive(self)

    def onRequest(self, auth: bool) -> bool:
        return False

    def onReport(self, engineID: bytes) -> bool:
        self.manager.nextState = TrustEveryResponse(self)
        return True

    def onResponse(self, engineID: bytes, auth: bool) -> bool:
        errmsg = "Received a ResponsePDU in response to a discovery message"
        raise DiscoveryError(errmsg)

# User provided engineID and has sent at least one request, but the
# Manager has not yet received any communication from the remote engine.
class Synchronizing(State):
    def onInactive(self) -> None:
        self.manager.nextState = Unsynchronized(self)

    def onRequest(self, auth: bool) -> bool:
        return not auth

    def onReport(self, engineID: bytes) -> bool:
        if self.manager.engineID == engineID:
            self.manager.nextState = RequireAuthentication(self)
            return True
        else:
            return False

    def onResponse(self, engineID: bytes, auth: bool) -> bool:
        if auth or self.manager.engineID == engineID:
            self.manager.nextState = RequireAuthentication(self)
            return True
        else:
            return False

class TrustEveryResponse(State):
    def onInactive(self) -> None:
        pass

    def onRequest(self, auth: bool) -> bool:
        return True

    def onReport(self, engineID: bytes) -> bool:
        return False

    def onResponse(self, engineID: bytes, auth: bool) -> bool:
        if auth:
            self.manager.nextState = RequireAuthentication(self)

        return True

class RequireAuthentication(State):
    def onInactive(self) -> None:
        pass

    def onRequest(self, auth: bool) -> bool:
        return True

    def onReport(self, engineID: bytes) -> bool:
        return False

    def onResponse(self, engineID: bytes, auth: bool) -> bool:
        return auth

class SharedBool:
    def __init__(self, value: bool) -> None:
        self.value = value

    def __bool__(self) -> bool:
        return self.value

    def makeTrue(self) -> None:
        self.value = True

    def makeFalse(self) -> None:
        self.value = False

class RequestMessage(RequestHandle[SNMPv3Message]):
    USM_EXCEPTION_TYPES = {
        usmStatsUnsupportedSecLevels:   UnsupportedSecurityLevelReport,
        usmStatsUnknownUserNames:       UnknownUserNameReport,
        usmStatsWrongDigests:           WrongDigestReport,
        usmStatsDecryptionErrors:       DecryptionErrorReport,
    }

    def __init__(self, request: "Request", synchronized: SharedBool) -> None:
        self.request = weakref.proxy(request)
        self.messageID: Optional[int] = None
        self.callback: Optional[Callable[[int], None]] = None

        self.synchronized = synchronized

    def addCallback(self, func: Callable[[int], None], msgID: int) -> None:
        self.callback = func
        self.messageID = msgID

    def push(self, response: SNMPv3Message) -> None:
        assert response.scopedPDU is not None
        pdu = response.scopedPDU.pdu

        if isinstance(pdu, ReportPDU):
            securityLevel = response.header.flags.securityLevel
            varbind = pdu.variableBindings[0]

            oid = varbind.name
            if oid == usmStatsUnknownEngineIDsInstance:
                self.request.processUnknownEngineID(response.securityEngineID)
            elif oid == usmStatsNotInTimeWindowsInstance:
                if not self.synchronized:
                    engineID = response.securityEngineID
                    self.request.processNotInTimeWindow(engineID)
            else:
                exception: UnhandledReport

                try:
                    exc_type = self.USM_EXCEPTION_TYPES[oid]
                except KeyError:
                    exception = UnrecognizedReport(varbind)
                else:
                    exception = exc_type(response, varbind)

                auth = response.header.flags.authFlag
                self.request.processException(exception, auth)
        else:
            self.request.processResponse(cast(ResponsePDU, response))

    def close(self) -> None:
        if self.callback is not None:
            assert self.messageID is not None
            self.callback(self.messageID)
            self.callback = None

class Request:
    def __init__(self,
        pdu: AnyPDU,
        manager: "SNMPv3UsmManager[Any]",
        userName: bytes,
        securityLevel: SecurityLevel,
        timeout: float = 10.0,
        refreshPeriod: float = 1.0,
    ) -> None:
        now = time.time()

        self._engineID: Optional[bytes] = None
        self.synchronized = SharedBool(False)

        self.manager = manager
        self.pdu = pdu
        self.securityLevel = securityLevel
        self.userName = userName

        self.messages: Set[RequestMessage] = set()

        self.event = threading.Event()
        self.exception: Optional[UnhandledReport] = None
        self.response: Optional[SNMPv3Message] = None

        self.expiration = now + timeout
        self._nextRefresh = self.expiration
        self.period = refreshPeriod

    def __del__(self) -> None:
        self.close()

    @property
    def engineID(self) -> Optional[bytes]:
        return self._engineID

    @engineID.setter
    def engineID(self, engineID: Optional[bytes]) -> None:
        if engineID == self._engineID:
            return

        if engineID is not None:
            self.manager.registerRemoteEngine(engineID)

        if self._engineID is not None:
            self.manager.unregisterRemoteEngine(self._engineID)

        self._engineID = engineID

    @property
    def expired(self) -> bool:
        return self.expiration <= time.time()

    @property
    def nextRefresh(self) -> float:
        return self._nextRefresh

    @nextRefresh.setter
    def nextRefresh(self, value: float) -> None:
        self._nextRefresh = min(self.expiration, value)

    def close(self) -> None:
        while self.messages:
            self.messages.pop().close()

        self.engineID = None

    def processNotInTimeWindow(self, engineID: bytes) -> None:
        self.manager.onReport(self, engineID)
        self.synchronized.makeTrue()
        self.synchronized = SharedBool(False)

    def processUnknownEngineID(self, engineID: bytes) -> None:
        if engineID != self.engineID:
            self.manager.onReport(self, engineID)

    def processException(self, exception: UnhandledReport, auth: bool) -> None:
        if auth:
            if not self.event.is_set():
                self.exception = exception
                self.event.set()
        elif self.exception is None:
            self.exception = exception

    def processResponse(self, response: SNMPv3Message) -> None:
        #if random.randint(1, 3) % 3 != 0:
        #    return

        if self.manager.processResponse(self, response):
            if not self.event.is_set():
                self.response = response
                self.event.set()

    # Always update self.nextRefresh right before calling this method
    def reallySend(self, engineID: Optional[bytes] = None) -> None:
        pdu = self.pdu
        user = self.userName
        securityLevel = self.securityLevel

        if engineID is None:
            if self.engineID is None:
                pdu = GetRequestPDU()
                engineID = b""
                user = b""
                securityLevel = noAuthNoPriv
            else:
                engineID = self.engineID
        else:
            self.engineID = engineID

        message = RequestMessage(self, self.synchronized)
        self.messages.add(message)
        self.manager.sendPdu(pdu, message, engineID, user, securityLevel)

    def refresh(self) -> Optional[float]:
        if self.event.is_set():
            return None

        now = time.time()
        timeToNextRefresh = self.nextRefresh - now

        if timeToNextRefresh <= 0.0:
            if self.expiration <= now:
                return None

            # Calculating it like this mitigates over-delay
            periodsElapsed = math.ceil(-timeToNextRefresh / self.period)
            self.nextRefresh += periodsElapsed * self.period
            self.reallySend()
            return 0.0
        else:
            return timeToNextRefresh

    def send(self, engineID: Optional[bytes] = None) -> None:
        now = time.time()
        self.nextRefresh = now + self.period
        self.reallySend(engineID)

    def wait(self) -> VarBindList:
        try:
            pdu = None
            while not self.expired:
                timeout = self.manager.refresh()

                try:
                    ready = self.event.wait(timeout=timeout)
                except KeyboardInterrupt as interrupt:
                    if self.exception is not None:
                        raise self.exception from interrupt
                    else:
                        raise

                if ready:
                    break

            if self.event.is_set() and self.response is not None:
                assert self.response.scopedPDU is not None
                pdu = cast(ResponsePDU, self.response.scopedPDU.pdu)
                if pdu.errorStatus:
                    raise ErrorResponse(
                        pdu.errorStatus,
                        pdu.errorIndex,
                        self.pdu,
                    )
                else:
                    return pdu.variableBindings
            else:
                if self.exception is not None:
                    raise self.exception
                else:
                    raise Timeout()
        finally:
            self.close()

T = TypeVar("T")
class SNMPv3UsmManager(Generic[T]):
    def __init__(self,
        dispatcher: Dispatcher[T],
        usm: UserBasedSecurityModule,
        channel: TransportChannel[T],
        namespace: str,
        defaultUserName: bytes,
        defaultSecurityLevel: SecurityLevel,
        engineID: Optional[bytes] = None,
        autowait: bool = True,
    ) -> None:

        # Used by @property
        self._engineID: Optional[bytes] = None

        # Read-only fields
        self.channel = channel
        self.namespace = namespace
        self.defaultUserName = defaultUserName
        self.defaultSecurityLevel = defaultSecurityLevel
        self.autowait = autowait

        self.dispatcher = dispatcher
        self.usm = usm

        self.generator = self.newGenerator()

        # protects self.active
        self.activeLock = threading.Lock()
        self.active: List[ComparableWeakRef[Request, float]] = []

        # protects self.unsent, self.state, and self.engineID
        self.lock = threading.Lock()
        self.unsent: Deque[ComparableWeakRef[Request, float]] = deque()

        self.state: State
        self.nextState = None
        if engineID is None:
            self.state = Inactive(self)
        else:
            self.engineID = engineID
            self.state = Unsynchronized(self)

    def __del__(self) -> None:
        self.engineID = None

    @property
    def engineID(self) -> Optional[bytes]:
        return self._engineID

    @engineID.setter
    def engineID(self, engineID: Optional[bytes]) -> None:
        if engineID == self._engineID:
            return

        if engineID is not None:
            self.registerRemoteEngine(engineID)

        if self._engineID is not None:
            self.unregisterRemoteEngine(self._engineID)

        self._engineID = engineID

    def generateRequestID(self) -> int:
        requestID = next(self.generator)

        if requestID == 0:
            self.generator = self.newGenerator()
            requestID = next(self.generator)

        return requestID

    def newGenerator(self) -> NumberGenerator:
        return NumberGenerator(32)

    def poke(self) -> bool:
        active = False
        with self.lock:
            self.state.onInactive()
            self.updateState()

            unsent = self.unsent
            self.unsent = deque()
            while unsent:
                reference = unsent.pop()
                request = reference()

                if request is not None:
                    if self.state.onRequest(request.securityLevel.auth):
                        with self.activeLock:
                            request.send(self.engineID)
                            heapq.heappush(self.active, reference)
                            active = True

                        self.updateState()
                    else:
                        self.unsent.appendleft(reference)

        return active

    def refresh(self) -> float:
        active = True
        while active:
            with self.activeLock:
                if self.active:
                    reference = self.active[0]
                    request = reference()

                    if request is None:
                        wait = None
                    else:
                        wait = request.refresh()

                    if wait is None:
                        heapq.heappop(self.active)
                        continue
                    elif wait > 0.0:
                        return wait
                    else:
                        heapq.heapreplace(self.active, reference)
                else:
                    active = False

            if not active:
                active = self.poke()

        return 0.0

    def registerRemoteEngine(self, engineID: bytes) -> None:
        if not self.usm.registerRemoteEngine(engineID, self.namespace):
            errmsg = f"Failed to register engineID {engineID!r}"

            if self.namespace:
                errmsg += f" under namespace \"{self.namespace}\""

            raise DiscoveryError(errmsg)

    def unregisterRemoteEngine(self, engineID: bytes) -> None:
        self.usm.unregisterRemoteEngine(engineID, self.namespace)

    def onReport(self, request: Request, engineID: bytes) -> None:
        with self.lock:
            sendAll = self.state.onReport(engineID)

            with self.activeLock:
                request.send(engineID)
                heapq.heapify(self.active)

            if sendAll:
                self.updateState()
                while self.unsent:
                    reference = self.unsent.pop()
                    req = reference()

                    if req is not None:
                        with self.activeLock:
                            req.send(engineID)
                            heapq.heappush(self.active, reference)

    def processResponse(self,
        request: Request,
        response: SNMPv3Message,
    ) -> bool:
        with self.lock:
            auth = response.header.flags.authFlag
            engineID = response.securityEngineID
            assert engineID is not None
            if self.state.onResponse(engineID, auth):
                self.engineID = engineID
                self.updateState()

            return True

    def sendPdu(self,
        pdu: AnyPDU,
        handle: RequestMessage,
        engineID: bytes,
        user: bytes,
        securityLevel: SecurityLevel,
    ) -> None:
        self.dispatcher.sendPdu(
            self.channel,
            ProtocolVersion.SNMPv3,
            pdu,
            handle,
            engineID,
            user,
            securityLevel=securityLevel,
            securityModel=SecurityModel.USM,
        )

    def sendRequest(self,
        pdu: AnyPDU,
        securityLevel: Optional[SecurityLevel] = None,
        user: Optional[bytes] = None,
        wait: Optional[bool] = None,
        **kwargs: Any,
    ) -> Union[Request, VarBindList]:
        if securityLevel is None:
            securityLevel = self.defaultSecurityLevel

        if user is None:
            user = self.defaultUserName

        if wait is None:
            wait = self.autowait

        pdu.requestID = self.generateRequestID()
        request = Request(pdu, self, user, securityLevel, **kwargs)
        reference = ComparableWeakRef(request, key=lambda r: r.nextRefresh)

        with self.lock:
            if self.state.onRequest(securityLevel.auth):
                with self.activeLock:
                    request.send(self.engineID)
                    heapq.heappush(self.active, reference)

                self.updateState()
            else:
                self.unsent.appendleft(reference)

        if wait:
            return request.wait()
        else:
            return request

    def updateState(self) -> None:
        if self.nextState is not None:
            self.state = self.nextState
            self.nextState = None

    def get(self,
        *oids: Union[str, OID],
        **kwargs: Any,
    ) -> Union[Request, VarBindList]:
        pdu = GetRequestPDU(*oids)
        return self.sendRequest(pdu, **kwargs)

    def getBulk(self,
        *oids: Union[str, OID],
        nonRepeaters: int = 0,
        maxRepetitions: int = 0,
        **kwargs: Any,
    ) -> Union[Request, VarBindList]:
        pdu = GetBulkRequestPDU(
            *oids,
            nonRepeaters=nonRepeaters,
            maxRepetitions=maxRepetitions,
        )

        return self.sendRequest(pdu, **kwargs)

    def getNext(self,
        *oids: Union[str, OID],
        **kwargs: Any,
    ) -> Union[Request, VarBindList]:
        pdu = GetNextRequestPDU(*oids)
        return self.sendRequest(pdu, **kwargs)

    def set(self,
        *varbinds: Union[
            Tuple[
                Union[str, OID],
                Optional[ASN1],
            ],
            VarBind,
        ],
        **kwargs: Any,
    ) -> Union[Request, VarBindList]:
        vbList = (VarBind(*varbind) for varbind in varbinds)
        pdu = SetRequestPDU(*vbList)
        return self.sendRequest(pdu, **kwargs)



__all__ = [
    "HmacMd5", "HmacSha",
    "HmacSha224", "HmacSha256", "HmacSha384", "HmacSha512"
]

import hashlib
import hmac

from snmp.security.usm import AuthProtocol
from snmp.typing import *

class HmacAuthProtocol(AuthProtocol):
    ALGORITHM:  ClassVar[Callable[..., "hashlib._Hash"]]
    N:          ClassVar[int]

    def __init__(self, key: bytes) -> None:
        self.key = key

    @classmethod
    def computeKey(cls, secret: bytes) -> bytes:
        repeat, truncate = divmod(1 << 20, len(secret))

        context = cls.ALGORITHM()
        for i in range(repeat):
            context.update(secret)

        context.update(secret[:truncate])
        return context.digest()

    @classmethod
    def localizeKey(cls, key: bytes, engineID: bytes) -> bytes:
        context = cls.ALGORITHM()
        context.update(key)
        context.update(engineID)
        context.update(key)
        return context.digest()

    @property
    def msgAuthenticationParameters(self) -> bytes:
        return bytes(self.N)

    def sign(self, data: bytes) -> bytes:
        context = hmac.new(self.key, digestmod=self.ALGORITHM)
        context.update(data)
        return context.digest()[:self.N]

class HmacMd5(HmacAuthProtocol):
    ALGORITHM = hashlib.md5
    N = 12

class HmacSha(HmacAuthProtocol):
    ALGORITHM = hashlib.sha1
    N = 12

class HmacSha224(HmacAuthProtocol):
    ALGORITHM = hashlib.sha224
    N = 16

class HmacSha256(HmacAuthProtocol):
    ALGORITHM = hashlib.sha256
    N = 24

class HmacSha384(HmacAuthProtocol):
    ALGORITHM = hashlib.sha384
    N = 32

class HmacSha512(HmacAuthProtocol):
    ALGORITHM = hashlib.sha512
    N = 48



__all__ = [
    "Engine",
    "SNMPv1", "SNMPv2c", "SNMPv3",
    "noAuthNoPriv", "authNoPriv", "authPriv",
]

from snmp.engine import Engine
from snmp.message import ProtocolVersion
from snmp.security.levels import *

SNMPv1  = ProtocolVersion.SNMPv1
SNMPv2c = ProtocolVersion.SNMPv2c
SNMPv3  = ProtocolVersion.SNMPv3



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['snmp',
 'snmp.manager',
 'snmp.message',
 'snmp.security',
 'snmp.security.usm',
 'snmp.security.usm.priv',
 'snmp.security.usm.priv.openssl',
 'snmp.security.usm.priv.pycryptodome',
 'snmp.transport']

package_data = \
{'': ['*'],
 'snmp': ['cffi/*', 'v3/*'],
 'snmp.transport': ['generic/*', 'posix/*']}

install_requires = \
['pycryptodome>=3.4']

setup(name='snmp',
      version='0.8.0',
      description='A user-friendly SNMP library',
      author=None,
      author_email='"Charles C. D. Tolley" <charlescdtolley@protonmail.com>',
      url=None,
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      python_requires='>=3.6',
     )
